"""Tests for astrality.actions.SymlinkAction."""

from pathlib import Path

from astrality.actions import SymlinkAction
from astrality.persistence import CreatedFiles


def test_null_object_pattern():
    """Copy actions without options should do nothing."""
    symlink_action = SymlinkAction(
        options={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    symlink_action.execute()


def test_symlink_dry_run(create_temp_files, caplog):
    """If dry_run is True, only log and not symlink."""
    content, target = create_temp_files(2)
    symlink_action = SymlinkAction(
        options={'content': str(content), 'target': str(target)},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    caplog.clear()
    result = symlink_action.execute(dry_run=True)

    # We should log the symlink that had been performed
    assert 'SKIPPED:' in caplog.record_tuples[0][2]
    assert str(content) in caplog.record_tuples[0][2]
    assert str(target) in caplog.record_tuples[0][2]

    # We should also still return the intended result
    assert result == {content: target}

    # But the symlink should not be created in a dry run
    assert not target.is_symlink()


def test_symlink_action_using_all_parameters(tmpdir):
    """All three parameters should be respected."""
    temp_dir = Path(tmpdir) / 'content'
    temp_dir.mkdir()

    target = Path(tmpdir) / 'target'
    target.mkdir()

    file1 = temp_dir / 'file1'
    file1.touch()

    file2 = temp_dir / 'file2'
    file2.touch()

    recursive_dir = temp_dir / 'recursive'
    recursive_dir.mkdir()

    file3 = temp_dir / 'recursive' / 'file3'
    file3.touch()

    symlink_options = {
        'content': str(temp_dir),
        'target': str(target),
        'include': r'file(\d)',
    }
    symlink_action = SymlinkAction(
        options=symlink_options,
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    symlink_action.execute()

    assert (target / '1').is_symlink()
    assert (target / '2').is_symlink()
    assert (target / 'recursive' / '3').is_symlink()

    assert (target / '1').resolve() == file1
    assert (target / '2').resolve() == file2
    assert (target / 'recursive' / '3').resolve() == file3


def test_symlinking_without_renaming(tmpdir):
    """When include is not given, keep symlink name."""
    temp_dir = Path(tmpdir) / 'content'
    temp_dir.mkdir()

    target = Path(tmpdir) / 'target'
    target.mkdir()

    file1 = temp_dir / 'file1'
    file1.touch()

    file2 = temp_dir / 'file2'
    file2.touch()

    recursive_dir = temp_dir / 'recursive'
    recursive_dir.mkdir()

    file3 = temp_dir / 'recursive' / 'file3'
    file3.touch()

    symlink_options = {
        'content': str(temp_dir),
        'target': str(target),
    }
    symlink_action = SymlinkAction(
        options=symlink_options,
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    symlink_action.execute()

    assert (target / 'file1').is_symlink()
    assert (target / 'file2').is_symlink()
    assert (target / 'recursive' / 'file3').is_symlink()

    assert (target / 'file1').resolve() == file1
    assert (target / 'file2').resolve() == file2
    assert (target / 'recursive' / 'file3').resolve() == file3


def test_symlinking_file_to_directory(tmpdir):
    """If symlinking from directory to file, place file in directory."""
    temp_dir = Path(tmpdir) / 'content'
    temp_dir.mkdir()

    target = Path(tmpdir) / 'target'
    target.mkdir()

    file1 = temp_dir / 'file1'
    file1.touch()

    symlink_options = {
        'content': str(file1),
        'target': str(target),
        'include': r'file1',
    }
    symlink_action = SymlinkAction(
        options=symlink_options,
        directory=temp_dir,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    symlink_action.execute()

    assert (target / 'file1').is_symlink()
    assert (target / 'file1').resolve() == file1
    assert symlink_action.symlinked_files == {
        file1: {target / 'file1'},
    }


def test_running_symlink_action_twice(create_temp_files):
    """Symlink action should be idempotent."""
    content, target = create_temp_files(2)
    content.write_text('content')
    target.write_text('target')

    symlink_options = {
        'content': str(content),
        'target': str(target),
    }
    symlink_action = SymlinkAction(
        options=symlink_options,
        directory=content.parent,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    # Symlink first time
    symlink_action.execute()
    assert target.is_symlink()
    assert target.read_text() == 'content'

    # A backup shoud be created
    backup = CreatedFiles().creations['test'][str(target)]['backup']
    assert Path(backup).read_text() == 'target'

    # Symlink one more time, and assert idempotency
    symlink_action.execute()
    assert target.is_symlink()
    assert target.read_text() == 'content'

    backup = CreatedFiles().creations['test'][str(target)]['backup']
    assert Path(backup).read_text() == 'target'


def test_backup_of_symlink_target(create_temp_files):
    """Overwritten copy targets should be backed up."""
    target, content = create_temp_files(2)

    # This file is the original and should be backed up
    target.write_text('original')

    # This is the new content which will be symlinked to
    content.write_text('new')

    symlink_options = {
        'content': str(content.name),
        'target': str(target),
    }
    symlink_action = SymlinkAction(
        options=symlink_options,
        directory=content.parent,
        replacer=lambda x: x,
        context_store={},
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    # We replace the content by executing the action
    symlink_action.execute()
    assert target.resolve().read_text() == 'new'

    # And when cleaning up the module, the backup should be restored
    CreatedFiles().cleanup(module='test')
    assert target.read_text() == 'original'


def test_cleanup_of_created_directory(create_temp_files, tmpdir):
    """Created directories should be cleaned up."""
    tmpdir = Path(tmpdir)
    [content] = create_temp_files(1)

    # The target requires a new directory to be created
    directory = tmpdir / 'dir'
    target = directory / 'target.tmp'

    # Execute the symlink action
    symlink_options = {
        'content': str(content.name),
        'target': str(target),
    }
    created_files = CreatedFiles().wrapper_for(module='test')
    symlink_action = SymlinkAction(
        options=symlink_options,
        directory=content.parent,
        replacer=lambda x: x,
        context_store={},
        creation_store=created_files,
    )
    symlink_action.execute()

    # The directory should now exist and be persisted
    assert directory.is_dir()
    assert directory in created_files.creation_store

    # But it should be deleted on cleanup
    CreatedFiles().cleanup(module='test')
    assert not directory.is_dir()



"""Tests for keep_running property of Module."""

from pathlib import Path

from astrality.module import Module, ModuleManager
from astrality.tests.utils import Retry


def test_module_that_does_not_need_to_keep_running():
    """on_startup block and on_exit block does not need to keep running."""
    module = Module(
        name='test',
        module_config={
            'run': {'shell': 'hi!'},
            'on_exit': {
                'run': {'shell': 'modified!'},
            },
        },
        module_directory=Path('/'),
    )
    assert module.keep_running is False


def test_module_that_needs_to_keep_running_due_to_on_modified():
    """on_modified block needs to keep running."""
    module = Module(
        name='test',
        module_config={
            'on_modified': {
                'some/path': {
                    'run': {'shell': 'modified!'},
                },
            },
        },
        module_directory=Path('/'),
    )
    assert module.keep_running is True


def test_module_that_needs_to_keep_running_due_to_on_event():
    """on_event with event listener need to keep running."""
    module = Module(
        name='test',
        module_config={
            'event_listener': {'type': 'weekday'},
            'on_event': {
                'run': {'shell': 'modified!'},
            },
        },
        module_directory=Path('/'),
    )
    assert module.keep_running is True


def test_module_without_event_listener_does_not_need_to_keep_running():
    """on_event without event listener need not to keep running."""
    module = Module(
        name='test',
        module_config={
            'on_event': {
                'run': {'shell': 'modified!'},
            },
        },
        module_directory=Path('/'),
    )
    assert module.keep_running is False


def test_that_reprocess_modified_files_causes_keep_running():
    """ModuleManager with reprocess_modified_files causes keep_running."""
    module_manager = ModuleManager(
        config={
            'modules': {
                'reprocess_modified_files': True,
            },
        },
    )
    assert module_manager.keep_running is True


def test_that_no_reprocess_modified_files_does_not_cause_keep_running():
    """ModuleManager without reprocess_modified_files does not keep_running."""
    module_manager = ModuleManager(
        config={
            'modules': {
                'reprocess_modified_files': False,
                'enabled_modules': [{'name': 'A'}],
            },
        },
        modules={'A': {}},
    )

    # We have to retry here, as processes from earlier tests might interfer
    assert Retry()(lambda: module_manager.keep_running is False)


def test_that_module_manager_asks_its_modules_if_it_should_keep_running():
    """ModuleManager should query its modules."""
    module_manager = ModuleManager(
        modules={
            'A': {'on_modified': {'some/path': {}}},
        },
    )
    assert module_manager.keep_running is True


def test_that_running_processes_causes_keep_running():
    """If shell commands are running, keep_running should be True."""
    module_manager = ModuleManager(
        modules={
            'A': {'run': {'shell': 'sleep 10'}},
        },
    )
    module_manager.finish_tasks()
    assert module_manager.keep_running is True



"""Tests for setup block in module."""

from pathlib import Path

from astrality.actions import SetupActionBlock
from astrality.module import Module


def test_that_module_block_is_persisted():
    """Module should create a 'setup' action block."""
    module_config = {
        'on_setup': {
            'run': {
                'shell': 'echo first time!',
            },
        },
    }
    params = {
        'name': 'test',
        'module_config': module_config,
        'module_directory': Path(__file__).parent,
    }

    module = Module(**params)
    assert isinstance(
        module.get_action_block(name='on_setup'),
        SetupActionBlock,
    )
    assert module.execute(action='run', block='on_setup') == (
        ('echo first time!', 'first time!'),
    )

    # After creating this module again, the run action should not be performed.
    del module
    module = Module(**params)
    assert module.execute(action='run', block='on_setup') == tuple()



"""Tests for the time_of_day event listener."""

from datetime import datetime

import pytest

from astrality.event_listener import TimeOfDay


@pytest.fixture
def default_time_of_day_event_listener():
    """Return a default time_of_day event listener object."""
    return TimeOfDay({'type': 'time_of_day'})


def test_processing_of_time_of_day_config(default_time_of_day_event_listener):
    """Test that the processing of the time_of_day config is correct."""
    assert 'sunday' not in default_time_of_day_event_listener.workdays
    assert 'monday' in default_time_of_day_event_listener.workdays

    assert default_time_of_day_event_listener.workdays['monday'].start.tm_hour \
        == 9
    assert default_time_of_day_event_listener.workdays['monday'].start.tm_min \
        == 0

    assert default_time_of_day_event_listener.workdays['friday'].end.tm_hour \
        == 17
    assert default_time_of_day_event_listener.workdays['friday'].end.tm_min \
        == 0


def test_current_event_of_time_of_day_event_listener(
    default_time_of_day_event_listener,
    freezer,
):
    """Test that the correct events are returned."""
    work_monday = datetime(year=2018, month=2, day=12, hour=10)
    freezer.move_to(work_monday)
    assert default_time_of_day_event_listener.event() == 'on'

    monday_freetime = datetime(year=2018, month=2, day=12, hour=18)
    freezer.move_to(monday_freetime)
    assert default_time_of_day_event_listener.event() == 'off'

    saturday = datetime(year=2018, month=2, day=17, hour=10)
    freezer.move_to(saturday)
    assert default_time_of_day_event_listener.event() == 'off'


def test_time_until_next_event_for_time_of_day_event_listener(
    default_time_of_day_event_listener,
    freezer,
):
    """Test that the correct number of seconds until next period is correct."""
    work_monday = datetime(year=2018, month=2, day=12, hour=10)
    freezer.move_to(work_monday)
    assert default_time_of_day_event_listener\
        .time_until_next_event()\
        .total_seconds() == 7 * 60 * 60 + 60

    monday_freetime = datetime(year=2018, month=2, day=12, hour=18)
    freezer.move_to(monday_freetime)
    assert default_time_of_day_event_listener\
        .time_until_next_event()\
        .total_seconds() == 15 * 60 * 60 + 60

    saturday = datetime(year=2018, month=2, day=17, hour=10)
    freezer.move_to(saturday)
    assert default_time_of_day_event_listener\
        .time_until_next_event()\
        .total_seconds() == 47 * 60 * 60 + 60



"""Tests for Module class."""

import logging
import os
from datetime import datetime, timedelta
from pathlib import Path

from freezegun import freeze_time
import pytest

from astrality import event_listener
from astrality.module import Module, ModuleManager
from astrality.context import Context
from astrality.tests.utils import RegexCompare, Retry


@pytest.fixture
def valid_module_section():
    return {
        'test_module': {
            'enabled': True,
            'event_listener': {'type': 'weekday'},
            'on_startup': {
                'run': [{'shell': 'echo {event}'}],
                'compile': [
                    {
                        'content': '../templates/test_template.conf',
                        'target': '/tmp/compiled_result',
                    },
                ],
            },
            'on_event': {
                'run': [{'shell': 'echo {../templates/test_template.conf}'}],
            },
            'on_exit': {
                'run': [{'shell': 'echo exit'}],
            },
        },
    }


@pytest.fixture
def simple_application_config():
    # Increase run timeout, so that we can inspect the shell results
    return {'config/modules': {'run_timeout': 2}}


@pytest.fixture
def module(
    valid_module_section,
    test_config_directory,
):
    return Module(
        name='test_module',
        module_config=valid_module_section['test_module'],
        module_directory=test_config_directory,
        context_store=Context({
            'fonts': {1: 'FuraMono Nerd Font'},
        }),
    )


@pytest.fixture
def single_module_manager(simple_application_config, valid_module_section):
    return ModuleManager(
        config=simple_application_config,
        modules=valid_module_section,
        context=Context({
            'fonts': {1: 'FuraMono Nerd Font'},
        }),
    )


class TestModuleClass:

    def test_valid_class_section_method_with_valid_section(
        self,
        valid_module_section,
    ):
        assert Module.valid_module(
            name='test',
            config=valid_module_section,
            requires_timeout=2,
            requires_working_directory=Path('/'),
        ) is True

    def test_valid_class_section_method_with_disabled_module_section(self):
        disabled_module_section = {
            'enabled': False,
            'on_startup': {'run': ['test']},
            'on_event': {'run': ['']},
            'on_exit': {'run': ['whatever']},
        }
        assert Module.valid_module(
            name='test',
            config=disabled_module_section,
            requires_timeout=2,
            requires_working_directory=Path('/'),
        ) is False

    def test_module_event_listener_class(self, module):
        assert isinstance(module.event_listener, event_listener.Weekday)

    def test_using_default_static_event_listener_when_no_event_listener_given(
        self,
        test_config_directory,
    ):
        static_module = Module(
            name='static',
            module_config={},
            module_directory=test_config_directory,
        )
        assert isinstance(static_module.event_listener, event_listener.Static)

    @freeze_time('2018-01-27')
    def test_running_module_manager_commands_with_special_interpolations(
        self,
        single_module_manager,
        caplog,
    ):
        single_module_manager.startup()
        assert (
            'astrality.actions',
            logging.INFO,
            'Running command "echo saturday".',
        ) in caplog.record_tuples
        assert (
            'astrality.utils',
            logging.INFO,
            'saturday',
        ) in caplog.record_tuples

        caplog.clear()
        single_module_manager.execute(
            action='run',
            block='on_event',
            module=single_module_manager.modules['test_module'],
        )
        assert (
            'astrality.actions',
            logging.INFO,
            RegexCompare(r'Running command "echo .+compiled_result"\.'),
        ) in caplog.record_tuples

    @freeze_time('2018-01-27')
    def test_running_module_startup_command(
        self,
        single_module_manager,
        module,
        valid_module_section,
        caplog,
    ):
        single_module_manager.startup()

        assert caplog.record_tuples == [
            (
                'astrality.compiler',
                logging.INFO,
                RegexCompare(
                    r'\[Compiling\].+test_template\.conf.+compiled_result"',
                ),
            ),
            (
                'astrality.actions',
                logging.INFO,
                'Running command "echo saturday".',
            ),
            (
                'astrality.utils',
                logging.INFO,
                'saturday',
            ),
        ]

    def test_running_module_on_event_command(
        self,
        single_module_manager,
        module,
        caplog,
    ):
        single_module_manager.startup()
        caplog.clear()

        single_module_manager.execute(
            action='run',
            block='on_event',
            module=single_module_manager.modules['test_module'],
        )

        # Convoluted way of getting the compilation target. Sorry!
        compiled_template = list(
            single_module_manager
            .modules['test_module']
            .performed_compilations()
            .values(),
        )[0].pop()

        for log_event in [
            (
                'astrality.actions',
                logging.INFO,
                f'Running command "echo {compiled_template}".',
            ),
            (
                'astrality.utils',
                logging.INFO,
                f'{compiled_template}',
            ),
        ]:
            assert log_event in caplog.record_tuples

    def test_running_module_exit_command(self, single_module_manager, caplog):
        single_module_manager.exit()
        for log_event in [
            (
                'astrality.actions',
                logging.INFO,
                'Running command "echo exit".',
            ),
            (
                'astrality.utils',
                logging.INFO,
                'exit',
            ),
        ]:
            assert log_event in caplog.record_tuples

    def test_missing_template_file(self, caplog):
        modules = {
            'test_module': {
                'on_startup': {
                    'compile': [
                        {'content': '/not/existing'},
                    ],
                },
            },
        }
        module_manager = ModuleManager(modules=modules)

        caplog.clear()
        module_manager.finish_tasks()
        assert 'Could not compile template "/not/existing" '\
               'to target "' in caplog.record_tuples[0][2]

    def test_compilation_of_template(
        self,
        valid_module_section,
        simple_application_config,
        module,
        conf,
        caplog,
    ):
        valid_module_section[
            'test_module'
        ][
            'event_listener'
        ][
            'type'
        ] = 'solar'

        compiled_template_content = 'some text\n' + os.environ['USER'] \
            + '\nFuraMono Nerd Font'
        module_manager = ModuleManager(
            config=simple_application_config,
            modules=valid_module_section,
            context=Context({
                'fonts': {1: 'FuraMono Nerd Font'},
            }),
        )
        directory = module_manager.config_directory

        caplog.clear()
        module_manager.execute(action='compile', block='on_startup')

        template_file = str(
            (directory / '../templates/test_template.conf').resolve(),
        )
        compiled_template = str(
            list(
                module_manager.modules['test_module']
                .performed_compilations()[Path(template_file)],
            )[0],
        )

        with open('/tmp/compiled_result', 'r') as file:
            compiled_result = file.read()

        assert compiled_template_content == compiled_result
        assert (
            'astrality.compiler',
            logging.INFO,
            f'[Compiling] Template: "{template_file}" '
            f'-> Target: "{compiled_template}"',
        ) in caplog.record_tuples


def test_running_finished_tasks_command(
    simple_application_config,
    valid_module_section,
    freezer,
    caplog,
):
    """Test that every task is finished at first finish_tasks() invocation."""
    thursday = datetime(
        year=2018,
        month=2,
        day=15,
        hour=12,
    )
    freezer.move_to(thursday)
    module_manager = ModuleManager(
        simple_application_config,
        modules=valid_module_section,
        context=Context({
            'fonts': {1: 'FuraMono Nerd Font'},
        }),
    )

    caplog.clear()
    module_manager.finish_tasks()

    # Only startup commands should be finished at first
    assert caplog.record_tuples == [
        (
            'astrality.compiler',
            logging.INFO,
            RegexCompare(
                r'\[Compiling\] Template: ".+/templates/test_template.conf" '
                r'-> Target: ".*compiled_result"',
            ),
        ),
        (
            'astrality.actions',
            logging.INFO,
            'Running command "echo thursday".',
        ),
        (
            'astrality.utils',
            logging.INFO,
            'thursday',
        ),
    ]

    # Now move one day ahead, and observe if event commands are run
    caplog.clear()
    friday = datetime(
        year=2018,
        month=2,
        day=16,
        hour=12,
    )
    freezer.move_to(friday)
    module_manager.finish_tasks()
    assert caplog.record_tuples == [
        (
            'astrality.module',
            logging.INFO,
            '[module/test_module] New event "friday". '
            'Executing actions.',
        ),
        (
            'astrality.actions',
            logging.INFO,
            RegexCompare(r'Running command "echo .+compiled_result"\.'),
        ),
        (
            'astrality.utils',
            logging.INFO,
            RegexCompare(r'.+compiled_result'),
        ),
    ]


def test_has_unfinished_tasks(
    simple_application_config,
    valid_module_section,
    freezer,
):
    # Move time to midday
    midday = datetime.now().replace(hour=12, minute=0)
    freezer.move_to(midday)

    # At instanziation, the module should have unfinished tasks
    weekday_module = ModuleManager(
        config=simple_application_config,
        modules=valid_module_section,
        context=Context({
            'fonts': {1: 'FuraMono Nerd Font'},
        }),
    )
    assert weekday_module.has_unfinished_tasks() is True

    # After finishing tasks, there should be no unfinished tasks (duh!)
    weekday_module.finish_tasks()
    assert weekday_module.has_unfinished_tasks() is False

    # If we move the time forwards, but not to a new event, there should still
    # not be any unfinished tasks
    before_midnight = datetime.now().replace(hour=23, minute=59)
    freezer.move_to(before_midnight)
    assert weekday_module.has_unfinished_tasks() is False

    # But right after a event (new weekday), there should be unfinished
    # tasks
    two_minutes = timedelta(minutes=2)
    freezer.move_to(before_midnight + two_minutes)
    assert weekday_module.has_unfinished_tasks() is True

    # Again, after finishing tasks, there should be no unfinished tasks left
    weekday_module.finish_tasks()
    assert weekday_module.has_unfinished_tasks() is False


@pytest.fixture
def config_with_modules():
    context = Context()
    modules = {
        'solar_module': {
            'enabled': True,
            'event_listener': {
                'type': 'solar',
                'longitude': 0,
                'latitude': 0,
                'elevation': 0,
            },
            'templates': {
                'template_name': {
                    'content': 'astrality/tests/templates/test_template.conf',
                    'target': '/tmp/compiled_result',
                },
            },
            'on_startup': {
                'run': [{'shell': 'echo solar compiling {template_name}'}],
            },
            'on_event': {
                'run': [{'shell': 'echo solar {event}'}],
            },
            'on_exit': {
                'run': [{'shell': 'echo solar exit'}],
            },
        },
        'weekday_module': {
            'enabled': True,
            'event_listener': {'type': 'weekday'},
            'on_startup': {
                'run': [{'shell': 'echo weekday startup'}],
            },
            'on_event': {
                'run': [{'shell': 'echo weekday {event}'}],
            },
            'on_exit': {
                'run': [{'shell': 'echo weekday exit'}],
            },
        },
        'disabled_module': {
            'enabled': False,
            'event_listener': 'static',
        },
    }

    return modules, context


@pytest.fixture
def module_manager(config_with_modules):
    modules, context = config_with_modules
    return ModuleManager(modules=modules, context=context)


def test_import_sections_on_event(config_with_modules, freezer):
    modules, context = config_with_modules

    modules[
        'weekday_module'
    ]['on_event']['import_context'] = [{
        'to_section': 'week',
        'from_path': 'astrality/tests/templates/weekday.yml',
        'from_section': '{event}',
    }]

    modules.pop('solar_module')
    module_manager = ModuleManager(
        modules=modules,
        context=Context({
            'fonts': {1: 'FuraCode Nerd Font'},
        }),
        directory=Path(__file__).parents[3],
    )

    assert module_manager.application_context['fonts'] \
        == Context({1: 'FuraCode Nerd Font'})

    sunday = datetime(year=2018, month=2, day=4)
    freezer.move_to(sunday)
    module_manager.finish_tasks()

    # Startup does not count as a event, so no context has been imported
    assert module_manager.application_context == Context({
        'fonts': Context({1: 'FuraCode Nerd Font'}),
    })

    monday = datetime(year=2018, month=2, day=5)
    freezer.move_to(monday)
    module_manager.finish_tasks()

    # The event has now changed, so context should be imported
    assert module_manager.application_context == {
        'fonts': Context({1: 'FuraCode Nerd Font'}),
        'week': Context({'day': 'monday'}),
    }


def test_import_sections_on_startup(config_with_modules, freezer):
    modules, context = config_with_modules

    # Insert day the module was started into 'start day'
    modules[
        'weekday_module'
    ]['on_startup']['import_context'] = [{
        'to_section': 'start_day',
        'from_path': 'astrality/tests/templates/weekday.yml',
        'from_section': '{event}',
    }]

    # Insert the current day into 'day_now'
    modules[
        'weekday_module'
    ]['on_event']['import_context'] = [{
        'to_section': 'day_now',
        'from_path': 'astrality/tests/templates/weekday.yml',
        'from_section': '{event}',
    }]
    modules.pop('solar_module')
    module_manager = ModuleManager(
        modules=modules,
        context=Context({
            'fonts': {1: 'FuraCode Nerd Font'},
        }),
        directory=Path(__file__).parents[3],
    )

    # Before finishing tasks, no context sections are imported
    assert module_manager.application_context['fonts'] \
        == {1: 'FuraCode Nerd Font'}

    # Start module on a monday
    sunday = datetime(year=2018, month=2, day=4)
    freezer.move_to(sunday)
    module_manager.finish_tasks()
    assert module_manager.application_context == {
        'fonts': Context({1: 'FuraCode Nerd Font'}),
        'start_day': Context({'day': 'sunday'}),
    }

    # 'now_day' should now be added, but 'start_day' should remain unchanged
    monday = datetime(year=2018, month=2, day=5)
    freezer.move_to(monday)
    module_manager.finish_tasks()
    assert module_manager.application_context == {
        'fonts': Context({1: 'FuraCode Nerd Font'}),
        'start_day': Context({'day': 'sunday'}),
        'day_now': Context({'day': 'monday'}),
    }


class TestModuleManager:
    @pytest.mark.slow
    def test_invocation_of_module_manager_with_config(self, conf):
        ModuleManager(conf)

    @pytest.mark.slow
    def test_using_finish_tasks_on_example_configuration(
        self,
        conf,
        modules,
        context,
    ):
        module_manager = ModuleManager(
            config=conf,
            modules=modules,
            context=context,
        )
        module_manager.finish_tasks()

    def test_number_of_modules_instanziated_by_module_manager(
        self,
        module_manager,
    ):
        assert len(module_manager) == 2


def test_time_until_next_event_of_several_modules(
    config_with_modules,
    module_manager,
    freezer,
):
    modules, context = config_with_modules

    solar_event_listener = event_listener.Solar(modules)
    noon = solar_event_listener.location.sun()['noon']

    one_minute = timedelta(minutes=1)
    freezer.move_to(noon - one_minute)

    assert module_manager.time_until_next_event() == one_minute
    two_minutes_before_midnight = datetime.now().replace(hour=23, minute=58)
    freezer.move_to(two_minutes_before_midnight)

    assert module_manager.time_until_next_event().total_seconds() \
        == timedelta(minutes=2).total_seconds()


def test_detection_of_new_event_involving_several_modules(
    config_with_modules,
    freezer,
):
    modules, context = config_with_modules

    # Move time to right before noon
    solar_event_listener = event_listener.Solar(modules)
    noon = solar_event_listener.location.sun()['noon']
    one_minute = timedelta(minutes=1)
    freezer.move_to(noon - one_minute)
    module_manager = ModuleManager(
        modules=modules,
        context=context,
    )

    # All modules should now considered to have now events
    assert module_manager.has_unfinished_tasks() is True

    # Running on event method for all the event changed modules
    module_manager.finish_tasks()

    # After running these methods, they should all be reverted to not changed
    assert module_manager.has_unfinished_tasks() is False

    # Move time to right after noon
    freezer.move_to(noon + one_minute)

    # The solar event listener should now be considered to have been event
    # changed
    assert module_manager.has_unfinished_tasks() is True

    # Again, check if on_event() method makes them unchanged
    module_manager.finish_tasks()
    assert module_manager.has_unfinished_tasks() is False

    # Move time two days forwards
    two_days = timedelta(days=2)
    freezer.move_to(noon + two_days)

    # Now both event listeners should be considered to have new events
    assert module_manager.has_unfinished_tasks() is True


def test_that_shell_filter_is_run_from_config_directory(test_config_directory):
    shell_filter_template = Path(__file__).parents[1] \
        / 'templates' / 'shell_filter_working_directory.template'
    shell_filter_template_target = Path(
        '/tmp/astrality/shell_filter_working_directory.template',
    )
    modules = {
        'A': {
            'on_startup': {
                'compile': [
                    {
                        'content': str(shell_filter_template),
                        'target': str(shell_filter_template_target),
                    },
                ],
            },
        },
    }

    module_manager = ModuleManager(modules=modules)
    module_manager.execute(action='compile', block='on_startup')

    with open(shell_filter_template_target) as compiled:
        assert compiled.read() == str(test_config_directory)

    os.remove(shell_filter_template_target)


@pytest.yield_fixture
def two_test_file_paths():
    test_file1 = Path('/tmp/astrality/test_file_1')
    test_file2 = Path('/tmp/astrality/test_file_2')

    yield test_file1, test_file2

    # Cleanup files after test has been run (if they exist)
    if test_file1.is_file():
        os.remove(test_file1)
    if test_file2.is_file():
        os.remove(test_file2)


def test_that_only_startup_event_block_is_run_on_startup(
    two_test_file_paths,
    test_config_directory,
    freezer,
):
    thursday = datetime(
        year=2018,
        month=2,
        day=15,
        hour=12,
    )
    freezer.move_to(thursday)

    test_file1, test_file2 = two_test_file_paths
    modules = {
        'A': {
            'event_listener': {'type': 'weekday'},
            'on_startup': {
                'run': [{'shell': 'touch ' + str(test_file1)}],
            },
            'on_event': {
                'run': [{'shell': 'touch ' + str(test_file2)}],
            },
        },
    }

    module_manager = ModuleManager(
        modules=modules,
    )

    # Before call to finish_tasks, no actions should have been performed
    assert not test_file1.is_file() and not test_file2.is_file()

    # Now call finish_tasks for the first time, only startup event block should
    # be run
    module_manager.finish_tasks()

    retry = Retry()
    assert retry(lambda: test_file1.is_file())
    assert retry(lambda: not test_file2.is_file())


def test_trigger_event_module_action(test_config_directory):
    modules = {
        'A': {
            'event_listener': {'type': 'weekday'},
            'on_startup': {
                'trigger': [
                    {'block': 'on_event'},
                    {'block': 'on_exit'},
                    {'block': 'on_modified', 'path': 'templateA'},
                ],
                'run': [{'shell': 'echo startup'}],
            },
            'on_event': {
                'run': [{'shell': 'echo on_event'}],
                'import_context': [{
                    'from_path': 'context/mercedes.yml',
                    'from_section': 'car',
                }],
            },
            'on_exit': {
                'run': [{'shell': 'echo exit'}],
            },
            'on_modified': {
                'templateA': {
                    'run': [{'shell': 'echo modified.templateA'}],
                    'compile': [
                        {'content': 'templateA'},
                    ],
                },
            },
        },
    }
    module_manager = ModuleManager(
        config={'modules': {'enabled_modules': [{'name': 'A'}]}},
        modules=modules,
    )

    # Check that all run commands have been imported into startup block
    results = tuple(module_manager.modules['A'].execute(
        action='run',
        block='on_startup',
    ))
    assert results == (
        ('echo startup', 'startup'),
        ('echo on_event', 'on_event'),
        ('echo exit', 'exit'),
        ('echo modified.templateA', 'modified.templateA'),
    )

    # Check that all context section imports are available in startup block
    # module_manager.modules['A'].import_context('on_startup')
    module_manager.modules['A'].execute(
        action='import_context',
        block='on_startup',
    )
    # TODO: Find out why Hanoi is included here
    assert module_manager.application_context['car'] == {
        'manufacturer': 'Mercedes',
    }

    # Double check that the other sections are not affected
    results = module_manager.modules['A'].execute(
        action='run',
        block='on_event',
    )
    assert results == (('echo on_event', 'on_event'),)

    results = module_manager.modules['A'].execute(
        action='run',
        block='on_exit',
    )
    assert results == (('echo exit', 'exit'),)

    module_manager.modules['A'].execute(
        action='import_context',
        block='on_event',
    )
    # TODO: Find out why Hanoi context is included here
    assert module_manager.application_context['car'] == {
        'manufacturer': 'Mercedes',
    }


def test_not_using_list_when_specifiying_trigger_action(conf_path):
    modules = {
        'A': {
            'on_startup': {
                'trigger': {'block': 'on_event'},
            },
            'on_event': {
                'run': [{'shell': 'echo on_event'}],
            },
        },
    }

    module_manager = ModuleManager(
        modules=modules,
        directory=conf_path,
    )

    # Check that all run commands have been imported into startup block
    result = module_manager.modules['A'].execute(
        action='run',
        block='on_startup',
    )
    assert result == (
        ('echo on_event', 'on_event'),
    )


def test_defining_on_startup_block_at_root_indentation(caplog):
    """Root indentation actions should be promoted to on_startup."""
    # Test that no root actions are not affected
    module_config = {
        'on_startup': {
            'run': [{'shell': 'echo on_startup'}],
        },
    }
    assert Module.prepare_on_startup_block(
        module_name='test',
        module_config=module_config,
    ) == {
        'on_startup': {
            'run': [{'shell': 'echo on_startup'}],
        },
    }

    # Test that actions are moved into empty block
    module_config = {
        'run': [{'shell': 'touch stuff'}],
        'compile': {'source': 'some/path'},
    }
    assert Module.prepare_on_startup_block(
        module_name='test',
        module_config=module_config,
    ) == {
        'on_startup': {
            'run': [{'shell': 'touch stuff'}],
            'compile': {'source': 'some/path'},
        },
    }

    # Test that overwriting values are logged
    caplog.clear()
    module_config = {
        'run': [{'shell': 'echo overwritten'}],
        'on_startup': {
            'run': [{'shell': 'echo original'}],
        },
    }
    assert Module.prepare_on_startup_block(
        module_name='test',
        module_config=module_config,
    ) == {
        'on_startup': {
            'run': [{'shell': 'echo overwritten'}],
        },
    }
    assert caplog.record_tuples[0][1] == logging.ERROR

    # Test that we only have partial overwrites
    module_config = {
        'stow': [{'content': 'new_stow'}],
        'copy': {'content': 'new_copy'},
        'on_startup': {
            'run': [{'shell': 'run'}],
            'stow': {'content': 'old_stow'},
            'copy': {'content': 'old_copy'},
        },
        'on_exit': {},
    }
    assert Module.prepare_on_startup_block(
        module_name='test',
        module_config=module_config,
    ) == {
        'on_startup': {
            'run': [{'shell': 'run'}],
            'stow': [{'content': 'new_stow'}],
            'copy': {'content': 'new_copy'},
        },
        'on_exit': {},
    }

    # Test that prepare_on_startup_block is actually used
    module_config = {
        'run': [{'shell': 'echo overwritten'}],
        'on_startup': {
            'run': [{'shell': 'echo original'}],
        },
    }
    module = Module(
        name='test_module',
        module_config=module_config,
        module_directory=Path('/'),
    )
    assert module.execute(action='run', block='on_startup') \
        == (('echo overwritten', 'overwritten'),)



"""Tests for requirements module."""

from pathlib import Path

from astrality.requirements import Requirement
from astrality.module import Module


def test_null_object_pattern():
    """Empty requirements should be considered satisfied."""
    successful_shell_requirement = Requirement(
        requirements={},
        directory=Path('/'),
    )
    assert successful_shell_requirement


def test_shell_command_requirement():
    """Requirement should be truthy when command returns 0 exit code."""
    successful_shell_requirement = Requirement(
        requirements={'shell': 'command -v ls'},
        directory=Path('/'),
    )
    assert successful_shell_requirement

    unsuccessful_shell_requirement = Requirement(
        requirements={'shell': 'command -v does_not_exist'},
        directory=Path('/'),
    )
    assert not unsuccessful_shell_requirement


def test_that_shell_commands_are_run_in_correct_diretory():
    """All shell commands should be run from 'directory'"""
    successful_shell_requirement = Requirement(
        requirements={'shell': 'ls tmp'},
        directory=Path('/'),
    )
    assert successful_shell_requirement

    unsuccessful_shell_requirement = Requirement(
        requirements={'shell': 'ls does_not_exist'},
        directory=Path('/'),
    )
    assert not unsuccessful_shell_requirement


def test_shell_command_timeout():
    """Shell commands can time out."""
    default_times_out = Requirement(
        requirements={'shell': 'sleep 0.1'},
        directory=Path('/'),
        timeout=0.01,
    )
    assert not default_times_out

    default_does_not_timeout = Requirement(
        requirements={'shell': 'sleep 0.1'},
        directory=Path('/'),
        timeout=0.2,
    )
    assert default_does_not_timeout

    specifed_does_not_timeout = Requirement(
        requirements={'shell': 'sleep 0.1', 'timeout': 0.2},
        directory=Path('/'),
        timeout=0.5,
    )
    assert specifed_does_not_timeout

    specified_does_timeout = Requirement(
        requirements={'shell': 'sleep 0.1', 'timeout': 0.05},
        directory=Path('/'),
        timeout=1000,
    )
    assert not specified_does_timeout


def test_environment_variable_requirement():
    """Requirement should be truthy when environment variable is available."""
    successful_env_requirement = Requirement(
        requirements={'env': 'EXAMPLE_ENV_VARIABLE'},
        directory=Path('/'),
    )
    assert successful_env_requirement

    unsuccessful_env_requirement = Requirement(
        requirements={'env': 'THIS_IS_NOT_SET'},
        directory=Path('/'),
    )
    assert not unsuccessful_env_requirement


def test_installed_requirement():
    """Requirement should be truthy when value is in $PATH."""
    successful_installed_requirement = Requirement(
        requirements={'installed': 'ls'},
        directory=Path('/'),
    )
    assert successful_installed_requirement

    unsuccessful_installed_requirement = Requirement(
        requirements={'installed': 'does_not_exist'},
        directory=Path('/'),
    )
    assert not unsuccessful_installed_requirement


def test_requiring_a_global_module():
    """You should be able to require global modules."""
    moduleA = Module(
        name='A',
        module_config={'requires': [{'module': 'B'}]},
        module_directory=Path(__file__).parent,
    )
    assert moduleA.depends_on == ('B',)

    moduleB = Module(
        name='B',
        module_config={'run': {'shell': 'echo hi!'}},
        module_directory=Path(__file__).parent,
    )
    assert moduleB.depends_on == tuple()

    moduleC = Module(
        name='C',
        module_config={'requires': [{'module': 'D'}]},
        module_directory=Path(__file__).parent,
    )
    assert moduleC.depends_on == ('D',)

    assert Requirement.pop_missing_module_dependencies(
        modules={'A': moduleA, 'B': moduleB, 'C': moduleC},
    ) == {'A': moduleA, 'B': moduleB}


def test_recursive_module_requirements():
    """Missing dependencies should propagate."""
    moduleA = Module(
        name='A',
        module_config={'requires': [{'module': 'B'}]},
        module_directory=Path(__file__).parent,
    )
    moduleB = Module(
        name='B',
        module_config={'requires': [{'module': 'C'}]},
        module_directory=Path(__file__).parent,
    )
    moduleC = Module(
        name='C',
        module_config={'requires': [{'module': 'D'}]},
        module_directory=Path(__file__).parent,
    )

    assert Requirement.pop_missing_module_dependencies(
        modules={'A': moduleA, 'B': moduleB, 'C': moduleC},
    ) == {}



import logging
import os
import psutil
import signal
import subprocess
import time

import pytest

from astrality.astrality import main, kill_old_astrality_processes
from astrality import utils
from astrality.tests.utils import Retry
from astrality.xdg import XDG


@pytest.mark.slow
def test_termination_of_main_process():
    astrality_process = subprocess.Popen(
        ['./bin/astrality'],
        stdout=subprocess.PIPE,
        preexec_fn=os.setsid,
    )

    time.sleep(2)
    os.killpg(os.getpgid(astrality_process.pid), signal.SIGTERM)

    astrality_process.wait()
    assert astrality_process.returncode == 0


@pytest.mark.slow
def test_interrupt_of_main_process():
    astrality_process = subprocess.Popen(
        ['./bin/astrality'],
        stdout=subprocess.PIPE,
        preexec_fn=os.setsid,
    )

    time.sleep(2)
    os.killpg(os.getpgid(astrality_process.pid), signal.SIGINT)

    astrality_process.wait()
    assert astrality_process.returncode == 0


def test_enabling_specific_module_from_command_line(
    caplog,
    monkeypatch,
    test_config_directory,
):
    """Modules parameter to main should enable specific module(s)."""
    monkeypatch.setitem(
        os.environ,
        'ASTRALITY_CONFIG_HOME',
        str(test_config_directory),
    )
    main(modules=['../test_modules/two_modules::bangladesh'], test=True)

    assert (
        'astrality.utils',
        logging.INFO,
        'Greetings from Dhaka!',
    ) in caplog.record_tuples


class TestKillOldAstralityProcesses:
    """Tests for astrality.astrality.kill_old_astrality_processes."""

    def test_killing_old_running_process(self):
        """The same running process should be killed."""
        perpetual_process = psutil.Popen([
            'python',
            '-c',
            '"from time import sleep; sleep(9999999999999)"',
        ])
        pidfile = XDG().data('astrality.pid')
        utils.dump_yaml(
            data=perpetual_process.as_dict(
                attrs=['pid', 'create_time', 'username'],
            ),
            path=pidfile,
        )
        kill_old_astrality_processes()
        assert Retry()(lambda: not perpetual_process.is_running())

    def test_not_killing_new_procces_with_same_pid(self):
        """The process should not be killed when it is not the original saved"""
        perpetual_process = psutil.Popen([
            'python',
            '-c',
            '"from time import sleep; sleep(9999999999999)"',
        ])

        process_data = perpetual_process.as_dict(
            attrs=['pid', 'create_time', 'username'],
        )
        process_data['create_time'] += 1

        utils.dump_yaml(
            data=process_data,
            path=XDG().data('astrality.pid'),
        )
        kill_old_astrality_processes()
        assert Retry()(lambda: perpetual_process.is_running())
        perpetual_process.kill()

    def test_trying_to_kill_process_no_longer_running(self):
        """No longer running processes should be handled gracefully."""
        finished_process = psutil.Popen(['echo', 'Done!'])
        process_data = finished_process.as_dict(
            attrs=['pid', 'create_time', 'username'],
        )
        finished_process.wait()

        utils.dump_yaml(
            data=process_data,
            path=XDG().data('astrality.pid'),
        )
        kill_old_astrality_processes()

    def test_killing_processes_when_no_previous_command_has_been_run(self):
        """The first ever invocation of the function should be handled."""
        pidfile = XDG().data('astrality.pid')
        pidfile.unlink()
        kill_old_astrality_processes()
        assert utils.load_yaml(
            path=pidfile,
        ) == psutil.Process().as_dict(
            attrs=['pid', 'create_time', 'username'],
        )



"""Tests for ImportContextAction class."""

from pathlib import Path

from astrality.actions import ImportContextAction
from astrality.context import Context
from astrality.persistence import CreatedFiles


def test_null_object_pattern():
    """Test initializing action with no behaviour."""
    import_context_action = ImportContextAction(
        options={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store=Context(),
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    import_context_action.execute()


def test_importing_entire_file(context_directory):
    """
    Test importing all sections from context file.

    All context sections should be imported in the absence of `from_section`.
    """
    context_import_dict = {
        'from_path': 'several_sections.yml',
    }
    context_store = Context()
    import_context_action = ImportContextAction(
        options=context_import_dict,
        directory=context_directory,
        replacer=lambda x: x,
        context_store=context_store,
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    import_context_action.execute()

    expected_context = {
        'section1': {
            'k1_1': 'v1_1',
            'k1_2': 'v1_2',
        },
        'section2': {
            'k2_1': 'v2_1',
            'k2_2': 'v2_2',
        },
    }
    assert context_store == expected_context


def test_importing_specific_section(context_directory):
    """Test importing specific sections from context file."""
    context_import_dict = {
        'from_path': 'several_sections.yml',
        'from_section': 'section1',
    }
    context_store = Context({'original': 'value'})
    import_context_action = ImportContextAction(
        options=context_import_dict,
        directory=context_directory,
        replacer=lambda x: x,
        context_store=context_store,
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    import_context_action.execute()

    expected_context = Context({
        'original': 'value',
        'section1': {
            'k1_1': 'v1_1',
            'k1_2': 'v1_2',
        },
    })
    assert context_store == expected_context


def test_replacer_function_being_used(context_directory):
    """
    Test use of replacement function in option retrieval.

    The function should be used when querying values from `options`.
    """
    context_import_dict = {
        'from_path': 'path',
        'from_section': 'from',
        'to_section': 'to',
    }
    context_store = Context()

    def replacer(option: str) -> str:
        if option == 'path':
            return 'several_sections.yml'
        elif option == 'from':
            return 'section1'
        elif option == 'to':
            return 'new_section'
        else:
            raise AssertionError

    import_context_action = ImportContextAction(
        options=context_import_dict,
        directory=context_directory,
        replacer=replacer,
        context_store=context_store,
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )
    import_context_action.execute()

    assert context_store == {
        'new_section': {
            'k1_1': 'v1_1',
            'k1_2': 'v1_2',
        },
    }


def test_that_replacer_is_run_every_time(context_directory):
    """
    The replacer should be run a new every time self.execute() is invoked.
    """
    context_import_dict = {
        'from_path': 'several_sections.yml',
        'from_section': 'section1',
        'to_section': 'whatever',
    }
    context_store = Context()

    class Replacer:
        def __init__(self) -> None:
            self.invoke_number = 0

        def __call__(self, option: str) -> str:
            self.invoke_number += 1
            return option

    replacer = Replacer()
    import_context_action = ImportContextAction(
        options=context_import_dict,
        directory=context_directory,
        replacer=replacer,
        context_store=context_store,
        creation_store=CreatedFiles().wrapper_for(module='test'),
    )

    import_context_action.execute()
    assert replacer.invoke_number == 3

    import_context_action.execute()
    assert replacer.invoke_number == 6



"""Tests for astrality.config.expand_path."""

from pathlib import Path

from astrality.config import expand_globbed_path, expand_path


def test_expand_path_method(test_config_directory):
    absolute_path = Path('/dir/ast')
    tilde_path = Path('~/dir')
    relative_path = Path('test')

    assert expand_path(
        path=absolute_path,
        config_directory=Path('/what/ever'),
    ) == absolute_path

    assert expand_path(
        path=tilde_path,
        config_directory=Path('/what/ever'),
    ) == Path.home() / 'dir'

    assert expand_path(
        path=relative_path,
        config_directory=test_config_directory,
    ) == test_config_directory / 'test'


def test_expansion_of_environment_variables(test_config_directory):
    """
    Environment variables should be expanded in paths.

    See pytest.ini for available environment variables.
    """
    assert expand_path(
        path=Path('${EXAMPLE_ENV_VARIABLE}/recursive'),
        config_directory=test_config_directory / '$EXAMPLE_ENV_VARIABLE',
    ) == test_config_directory / 'test_value' / 'test_value' / 'recursive'


def test_expand_globbed_path(test_config_directory):
    """Globbed paths should allow one level of globbing."""
    templates = Path('test_modules', 'using_all_actions')
    paths = expand_globbed_path(
        path=templates / '*',
        config_directory=test_config_directory,
    )
    assert len(paths) == 5
    assert test_config_directory / templates / 'module.template' in paths


def test_expand_recursive_globbed_path(test_config_directory):
    """Globbed paths should allow recursive globbing."""
    templates = Path('test_modules', 'using_all_actions')
    paths = expand_globbed_path(
        path=templates / '**' / '*',
        config_directory=test_config_directory,
    )
    assert len(paths) == 6
    assert test_config_directory / templates / 'recursive' / 'empty.template' \
        in paths



"""Module implementing user configured custom functionality."""

import logging
from collections import defaultdict
from datetime import timedelta
from pathlib import Path
import psutil
import re
from typing import (
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Match,
    Optional,
    Set,
    Tuple,
    Union,
)

from mypy_extensions import TypedDict

from astrality.actions import ActionBlock, ActionBlockDict, SetupActionBlock
from astrality.config import (
    AstralityYAMLConfigDict,
    GlobalModulesConfig,
    expand_path,
    user_configuration,
)
from astrality.event_listener import (
    EventListener,
    EventListenerConfig,
    event_listener_factory,
)
from astrality.filewatcher import DirectoryWatcher
from astrality.context import Context
from astrality.requirements import Requirement, RequirementDict
from astrality.utils import cast_to_list


class ModuleConfigDict(TypedDict, total=False):
    """Content of module configuration dict."""

    enabled: Optional[bool]
    requires: Union[RequirementDict, List[RequirementDict]]
    event_listener: EventListenerConfig

    on_setup: ActionBlockDict
    on_startup: ActionBlockDict
    on_exit: ActionBlockDict
    on_event: ActionBlockDict
    on_modified: Dict[str, ActionBlockDict]


class ModuleActionBlocks(TypedDict):
    """Contents of Module().action_blocks."""

    on_setup: SetupActionBlock
    on_startup: ActionBlock
    on_event: ActionBlock
    on_exit: ActionBlock
    on_modified: Dict[Path, ActionBlock]


logger = logging.getLogger(__name__)


class Module:
    """
    Class for determining module actions.

    A Module instance should only return actions to be performed, but never
    perform them. That is the responsibility of a ModuleManager instance.

    A module can define a set of commands to be run on astrality startup, and
    exit, in addition to every time a given type of event changes.

    Commands are run in the users shell, and can use the following placeholder:
    - {event}: The event specified by the event_listener instance.

    :param module_config: Dictionary keyed to module name and containing all
        module options.
    :param module_directory: The directory which contains all module relevant
        files, such as `config.yml`. All relative paths use this as anchor.
    :param replacer: String options should be processed by this function in
        order to replace relevant placeholders.
    :param global_modules_config: GlobalModulesConfig object specifying
        configuration options applicable to all modules, such as run_timeout.
    :param dry_run: If file system actions should be printed and skipped.
    """

    action_blocks: ModuleActionBlocks
    depends_on: Tuple[str]

    def __init__(
        self,
        name: str,
        module_config: ModuleConfigDict,
        module_directory: Path,
        replacer: Callable[[str], str] = lambda string: string,
        context_store: Context = Context(),
        global_modules_config: Optional[GlobalModulesConfig] = None,
        dry_run: bool = False,
    ) -> None:
        """
        Initialize Module object with a section from a config dictionary.

        Section name must be [module/*], where * is the module name.
        In addition, the enabled option must be set to "true", or not set at
        all.

        module_config example:
        {'name':
            'enabled': True,
            'event_listener': {'type': 'weekday'},
            'on_startup': {'run': ['echo weekday is {event}']},
        }
        """
        self.name = name

        # The source directory for the module, determining how to interpret
        # relative paths in the module config
        self.directory = module_directory

        # All user string options should be processed by the replacer
        self.replace = replacer

        # Use static event_listener if no event_listener is specified
        self.event_listener: EventListener = \
            event_listener_factory(
                module_config.get(
                    'event_listener',
                    {'type': 'static'},
                ),
            )

        self.context_store = context_store

        # Move root actions to 'on_startup' block
        module_config = self.prepare_on_startup_block(
            module_name=self.name,
            module_config=module_config,
        )

        # Create action block object for each available action block type
        action_blocks: ModuleActionBlocks = {'on_modified': {}}  # type: ignore
        params = {
            'module_name': self.name,
            'directory': self.directory,
            'replacer': self.interpolate_string,
            'context_store': self.context_store,
            'global_modules_config': global_modules_config,
        }

        # Create special case setup action block, it removes any action already
        # performed.
        action_blocks['on_setup'] = SetupActionBlock(  # type: ignore
            action_block=module_config.get('on_setup', {}),
            **params,
        )

        # Create normal action blocks
        for block_name in ('on_startup', 'on_event', 'on_exit'):
            action_blocks[block_name] = ActionBlock(  # type: ignore
                action_block=module_config.get(  # type: ignore
                    block_name,
                    {},
                ),
                **params,
            )

        for path_string, action_block_dict in module_config.get(
            'on_modified',
            {},
        ).items():
            modified_path = expand_path(
                path=Path(path_string),
                config_directory=self.directory,
            )
            action_blocks['on_modified'][modified_path] = \
                ActionBlock(  # type: ignore
                    action_block=action_block_dict,
                    **params,
            )

        self.action_blocks = action_blocks

        requirements = cast_to_list(module_config.get('requires', []))
        self.depends_on = tuple(  # type: ignore
            requirement['module']
            for requirement
            in requirements
            if 'module' in requirement
        )

    @staticmethod
    def prepare_on_startup_block(
        module_name: str,
        module_config: ModuleConfigDict,
    ) -> ModuleConfigDict:
        """
        Move actions at root indentation to on_startup block.

        NB! Mutates module_config in-place.

        :param module_config: Module configuration dictionary.
        :return: Configuration dictionary with root actions moved to on_startup.
        """
        action_types = ActionBlock.action_types.keys()
        root_level_action_types = [
            action_type
            for action_type
            in action_types
            if action_type in module_config
        ]

        if not root_level_action_types:
            return module_config

        if 'on_startup' in module_config:
            logger.error(
                f'[module/{module_name}] Actions defined both at root '
                'indentation and in "on_startup" block. This is reduntant! '
                'Root action types might overwrite "on_startup" actions.',
            )
        else:
            module_config['on_startup'] = {}

        for action_type in root_level_action_types:
            module_config[
                'on_startup'
            ][action_type] = module_config.pop(action_type)  # type: ignore

        return module_config

    def get_action_block(
        self,
        name: str,
        path: Optional[Path] = None,
    ) -> ActionBlock:
        """
        Return specific action block from module.

        :param name: Identifier of action block, for example 'on_startup'.
        :param path: If `name` == on_modified, path specifies modified path.
        """
        if path:
            assert path.is_absolute()
            assert name == 'on_modified'
            return self.action_blocks[name][path]  # type: ignore
        else:
            assert name in ('on_setup', 'on_startup', 'on_event', 'on_exit')
            return self.action_blocks[name]  # type: ignore

    def execute(
        self,
        action: str,
        block: str,
        path: Optional[Path] = None,
        dry_run: bool = False,
    ) -> Optional[Tuple[Tuple[str, str], ...]]:
        """
        Perform action defined in action block.

        :param actions: Action type to be performed such as 'compile'.
            If passed 'all' then all actions types will be triggered.
        :param block_name: Name of block such as 'on_startup'.
        :param path: Absolute path in case of block_name == 'on_modified'.
        :param dry_run: If external side-effects should be skipped.
        :return: Optional tuple of 2-tuples if run actions have been performed.
            First item is command being run, second item is the standard output
            of the shell command.
        """
        if action == 'all':
            # If 'all' is specified, then we can run all actions except trigger,
            # as triggers are handled in each respective action.
            all_actions = filter(
                lambda x: x != 'trigger',
                ActionBlock.action_types.keys(),
            )

            results: Tuple[Tuple[str, str], ...] = tuple()
            for action in all_actions:
                result = self.execute(
                    action=action,
                    block=block,
                    path=path,
                    dry_run=dry_run,
                )
                if result:
                    results += result

            return results

        # In this branch, we have a single action, for example 'run'.
        assert isinstance(action, str)
        assert action in ActionBlock.action_types
        action_block = self.get_action_block(name=block, path=path)
        results = getattr(action_block, action)(dry_run=dry_run)

        # We need to execute the same action in any triggered action block
        triggers = action_block.triggers()
        for trigger in triggers:
            result = self.execute(
                action=action,
                block=trigger.block,
                path=trigger.absolute_path,
                dry_run=dry_run,
            )
            if result:
                results += result

        return results

    def all_action_blocks(self) -> Iterable[ActionBlock]:
        """Return flatten tuple of all module action blocks."""
        return (
            self.action_blocks['on_setup'],
            self.action_blocks['on_startup'],
            self.action_blocks['on_event'],
            self.action_blocks['on_exit'],
            *self.action_blocks['on_modified'].values(),
        )

    def performed_compilations(self) -> DefaultDict[Path, Set[Path]]:
        """
        Return all templates that have been compiled and their target(s).

        :return: Dictionary with template path keys and values as a set of
            compilation target paths for that template.
        """
        performed_compilations: DefaultDict[Path, Set[Path]] = defaultdict(set)
        for action_block in self.all_action_blocks():
            for template, targets \
                    in action_block.performed_compilations().items():
                performed_compilations[template] |= targets

        return performed_compilations

    def interpolate_string(self, string: str) -> str:
        """
        Replace all module placeholders in string.

        The configuration string processor replaces {event} with the current
        module event, and {/path/to/template} with the compilation target.

        :return: String where '{path/to/template}' has been replaced with
            'path/to/compilation/target', and {event} repleced with last event.
        """
        # First replace any event placeholders with the last event, this must
        # be done before path replacements as paths could contain {event}.
        string = self.replace(
            string.replace(
                '{event}',
                self.event_listener.event(),
            ),
        )

        placeholder_pattern = re.compile(r'({.+})')
        performed_compilations = self.performed_compilations()

        def replace_placeholders(match: Match) -> str:
            """Regex file path match replacer."""
            # Remove enclosing curly brackets
            specified_path = match.group(0)[1:-1]

            absolute_path = expand_path(
                path=Path(specified_path),
                config_directory=self.directory,
            )

            if absolute_path in performed_compilations:
                # TODO: Is joining the right thing to do here?
                return ' '.join(
                    [
                        str(path)
                        for path
                        in performed_compilations[absolute_path]
                    ],
                )
            else:
                logger.error(
                    'String placeholder {' + specified_path + '} '
                    f'could not be replaced. "{specified_path}" '
                    'has not been compiled.',
                )
                # Return the placeholder left alone
                return '{' + specified_path + '}'

        return placeholder_pattern.sub(
            repl=replace_placeholders,
            string=string,
        )

    @property
    def keep_running(self) -> bool:
        """Return True if Module needs to keep running."""
        on_modifed = bool(self.action_blocks['on_modified'])
        event_listener = bool(self.action_blocks['on_event']) \
            and self.event_listener.type_ != 'static'

        return on_modifed or event_listener

    def __repr__(self) -> str:
        """Return string representation of Module object."""
        return f'Module(name={self.name})'

    @staticmethod
    def valid_module(
        name: str,
        config: ModuleConfigDict,
        requires_timeout: Union[int, float],
        requires_working_directory: Path,
    ) -> bool:
        """
        Check if the given dict represents a valid enabled module.

        The method determines this by inspecting any requires items the module
        might have specified.

        :param name: Name of module, used for logging purposes.
        :param config: Configuration dictionary of the module.
        :param requires_timeout: Time to wait for shell command requirements.
        :param requires_working_directory: CWD for shell commands.
        :return: True if module should be enabled.
        """
        if not config.get('enabled', True):
            return False

        # The module is enabled, now check if all requirements are satisfied
        requires: List[RequirementDict] = cast_to_list(
            config.get(
                'requires',
                {},
            ),
        )
        requirements = [
            Requirement(
                requirements=requirements_dict,
                directory=requires_working_directory,
                timeout=requires_timeout,
            )
            for requirements_dict
            in requires
        ]
        if all(requirements):
            return True
        else:
            logger.warning(
                f'[module/{name}] ' +
                ', '.join([
                    repr(requirement)
                    for requirement
                    in requirements
                ]) + '!',
            )
            return False


class ModuleManager:
    """
    A manager for operating on a set of modules.

    :param config: Global configuration options.
    :param modules: Dictionary containing globally defined modules.
    :param context: Global context.
    :param directory: Directory containing global configuration.
    :param dry_run: If file system actions should be printed and skipped.
    """

    def __init__(
        self,
        config: AstralityYAMLConfigDict = {},
        modules: Dict[str, ModuleConfigDict] = {},
        context: Context = Context(),
        directory: Path = Path(__file__).parent / 'tests' / 'test_config',
        dry_run: bool = False,
    ) -> None:
        """Initialize a ModuleManager object from `astrality.yml` dict."""
        self.config_directory = directory
        self.application_config = config
        self.application_context = context
        self.dry_run = dry_run

        self.startup_done = False
        self.last_module_events: Dict[str, str] = {}

        # Get module configurations which are externally defined
        self.global_modules_config = GlobalModulesConfig(
            config=config.get('modules', {}),
            config_directory=self.config_directory,
        )
        self.reprocess_modified_files = \
            self.global_modules_config.reprocess_modified_files

        self.modules: Dict[str, Module] = {}

        # Insert externally managed modules
        for external_module_source \
                in self.global_modules_config.external_module_sources:
            # Insert context defined in external configuration
            module_context = external_module_source.context(
                context=self.application_context,
            )
            self.application_context.reverse_update(module_context)

            module_configs = external_module_source.modules(
                context=self.application_context,
            )
            module_directory = external_module_source.directory

            for module_name, module_config in module_configs.items():
                if module_name \
                        not in self.global_modules_config.enabled_modules:
                    continue

                if not Module.valid_module(
                    name=module_name,
                    config=module_config,
                    requires_timeout=self.global_modules_config.
                    requires_timeout,  # noqa
                    requires_working_directory=module_directory,
                ):
                    continue

                module = Module(
                    name=module_name,
                    module_config=module_config,
                    module_directory=module_directory,
                    replacer=self.interpolate_string,
                    context_store=self.application_context,
                    global_modules_config=self.global_modules_config,
                    dry_run=dry_run,
                )
                self.modules[module.name] = module

        # Insert modules defined in `astrality.yml`
        for module_name, module_config in modules.items():
            # Check if this module should be included
            if module_name not in self.global_modules_config.enabled_modules:
                continue

            if not Module.valid_module(
                name=module_name,
                config=module_config,
                requires_timeout=self.global_modules_config.requires_timeout,
                requires_working_directory=self.config_directory,
            ):
                continue

            module = Module(
                name=module_name,
                module_config=module_config,
                module_directory=self.config_directory,
                replacer=self.interpolate_string,
                context_store=self.application_context,
                global_modules_config=self.global_modules_config,
                dry_run=dry_run,
            )
            self.modules[module.name] = module

        # Remove modules which depends on other missing modules
        Requirement.pop_missing_module_dependencies(self.modules)

        # Initialize the config directory watcher, but don't start it yet
        self.directory_watcher = DirectoryWatcher(
            directory=self.config_directory,
            on_modified=self.file_system_modified,
        )

        logger.info('Enabled modules: ' + ', '.join(self.modules.keys()))

    def module_events(self) -> Dict[str, str]:
        """Return dict containing the event of all modules."""
        module_events = {}
        for module_name, module in self.modules.items():
            module_events[module_name] = module.event_listener.event()

        return module_events

    def finish_tasks(self) -> None:
        """
        Finish all due tasks defined by the managed modules.

        The order of finishing tasks is as follows:
            1) Import any relevant context sections.
            2) Compile all templates with the new section.
            3) Run startup commands, if it is not already done.
            4) Run on_event commands, if it is not already done for this
               module events combination.
        """
        if not self.startup_done:
            # Save the last event configuration, such that on_event
            # is only run when the event *changes*
            self.last_module_events = self.module_events()

            # Perform setup actions not yet executed
            self.setup()

            # Perform all startup actions
            self.startup()
        elif self.last_module_events != self.module_events():
            # One or more module events have changed, execute the event blocks
            # of these modules.

            for module_name, event in self.module_events().items():
                if not self.last_module_events[module_name] == event:
                    logger.info(
                        f'[module/{module_name}] New event "{event}". '
                        'Executing actions.',
                    )
                    self.execute(
                        action='all',
                        block='on_event',
                        module=self.modules[module_name],
                    )
                    self.last_module_events[module_name] = event

    def has_unfinished_tasks(self) -> bool:
        """Return True if there are any module tasks due."""
        if not self.startup_done:
            return True
        else:
            return self.last_module_events != self.module_events()

    def time_until_next_event(self) -> timedelta:
        """Time left until first event change of any of the modules managed."""
        try:
            return min(
                module.event_listener.time_until_next_event()
                for module
                in self.modules.values()
            )
        except ValueError:
            return timedelta.max

    def execute(
        self,
        action: str,
        block: str,
        module: Optional[Module] = None,
    ) -> None:
        """
        Execute action(s) specified in managed modules.

        The module actions are executed according to their specified priority.
        First import context, then symlink, and so on...

        :param action: Action to be perfomed. If given 'all', then all actions
            will be performed.
        :param block: Action block to be executed, for example 'on_exit'.
        :module: Specific module to be executed. If not provided, then all
            managed modules will be executed.
        """
        assert block in ('on_setup', 'on_startup', 'on_event', 'on_exit')

        modules: Iterable[Module]
        if isinstance(module, Module):
            modules = (module, )
        else:
            modules = self.modules.values()

        if action == 'all':
            all_actions = filter(
                lambda x: x != 'trigger',
                ActionBlock.action_types.keys(),
            )
        else:
            all_actions = (action,)  # type: ignore

        for specific_action in all_actions:
            for module in modules:
                module.execute(
                    action=specific_action,
                    block=block,
                    dry_run=self.dry_run,
                )

    def setup(self) -> None:
        """
        Run setup actions specified by the managed modules, not yet executed.
        """
        self.execute(action='all', block='on_setup')

    def startup(self):
        """
        Run all startup actions specified by the managed modules.

        Also starts the directory watcher in $ASTRALITY_CONFIG_HOME.
        """
        assert not self.startup_done
        self.execute(action='all', block='on_startup')
        self.directory_watcher.start()
        self.startup_done = True

    def exit(self):
        """
        Run all exit tasks specified by the managed modules.

        Also close all temporary file handlers created by the modules.
        """
        self.execute(action='all', block='on_exit')

        # Stop watching config directory for file changes
        self.directory_watcher.stop()

    def on_modified(self, modified: Path) -> bool:
        """
        Perform actions when a watched file is modified.

        :return: Returns True if on_modified block was triggered.
        """
        assert modified.is_absolute()
        triggered = False

        for module in self.modules.values():
            if modified not in module.action_blocks['on_modified']:
                continue

            triggered = True
            logger.info(
                f'[module/{module.name}] on_modified:{modified} triggered.',
            )

            module.execute(
                action='all',
                block='on_modified',
                path=modified,
                dry_run=self.dry_run,
            )

        return triggered

    def file_system_modified(self, modified: Path) -> None:
        """
        Perform actions for when files within the config directory are modified.

        Run any context imports, compilations, and shell commands specified
        within the on_modified event block of each module.

        Also, if hot_reload is True, we reinstantiate the ModuleManager object
        if the application configuration has been modified.
        """
        config_files = (
            self.config_directory / 'astrality.yml',
            self.config_directory / 'modules.yml',
            self.config_directory / 'context.yml',
        )

        if modified in config_files:
            logger.info(
                f'$ASTRALITY_CONFIG_HOME/{modified.name} has been modified!',
            )
            self.on_application_config_modified()
            return
        else:
            # Run any relevant on_modified blocks.
            triggered = self.on_modified(modified)

        if not triggered:
            # Check if the modified path is a template which is supposed to
            # be recompiled.
            self.recompile_modified_template(modified=modified)

    def on_application_config_modified(self):
        """
        Reload the ModuleManager if astrality.yml has been modified.

        Reloadnig the module manager only occurs if the user has configured
        `hot_reload_config`.
        """
        if not self.application_config.get(
            'astrality',
            {},
        ).get(
            'hot_reload_config',
            False,
        ):
            # Hot reloading is not enabled, so we return early
            logger.info('"hot_reload" disabled.')
            return

        # Hot reloading is enabled, get the new configuration dict
        logger.info('Reloading $ASTRALITY_CONFIG_HOME...')
        (
            new_application_config,
            new_modules,
            new_context,
            directory,
        ) = user_configuration(
            config_directory=self.config_directory,
        )

        try:
            # Reinstantiate this object
            new_module_manager = ModuleManager(
                config=new_application_config,
                modules=new_modules,
                context=new_context,
                directory=directory,
            )

            # Run all old exit actions, since the new config is valid
            self.exit()

            # Swap place with the new configuration
            self = new_module_manager

            # Run startup commands from the new configuration
            self.finish_tasks()
        except Exception:
            # New configuration is invalid, just keep the old one
            # TODO: Test this behaviour
            logger.error('New configuration detected, but it is invalid!')
            pass

    def recompile_modified_template(self, modified: Path):
        """
        Recompile any modified template if configured.

        This requires setting the global setting:
        reprocess_modified_files: true
        """
        if not self.reprocess_modified_files:
            return

        # Run any compile action a new if that compile action uses the modifed
        # path as a template.
        for module in self.modules.values():
            for action_block in module.all_action_blocks():
                for compile_action in action_block._compile_actions:
                    if modified in compile_action:
                        compile_action.execute(dry_run=self.dry_run)

                for stow_action in action_block._stow_actions:
                    if modified in stow_action:
                        stow_action.execute(dry_run=self.dry_run)

                # TODO: Test this branch
                for copy_action in action_block._copy_actions:
                    if modified in copy_action:
                        copy_action.execute(dry_run=self.dry_run)

    def interpolate_string(self, string: str) -> str:
        """
        Process configuration string before using it.

        This function is passed as a reference to all modules, making them
        perform the replacement instead. For now, the ModuleManager does not
        change the string at all, but we will leave it here in case we would
        want to interpolate strings from the ModuleManager level instead of
        the Module level.

        :param string: String to be processed.
        :return: Processed string.
        """
        return string

    @property
    def keep_running(self) -> bool:
        """Return True if ModuleManager needs to keep running."""
        if self.reprocess_modified_files:
            return True

        if any(module.keep_running for module in self.modules.values()):
            return True

        current_process = psutil.Process()
        children = current_process.children(recursive=False)
        return bool(children)

    def __len__(self) -> int:
        """Return the number of managed modules."""
        return len(self.modules)

    def __del__(self) -> None:
        """Close filesystem watcher if enabled."""
        if hasattr(self, 'directory_watcher'):
            self.directory_watcher.stop()



"""Tests for astrality.persistence.CreatedFiles."""

from pathlib import Path
import shutil

import pytest

from astrality.persistence import (
    CreatedFiles,
    CreationMethod,
)


def test_that_file_is_created_with_created_files():
    """A file should be created to store module created files."""
    created_files = CreatedFiles()
    path = created_files.path
    assert path.name == 'created_files.yml'
    assert path.exists()


def test_that_created_files_are_properly_persisted(create_temp_files):
    """Inserted files should be persisted properly."""
    created_files = CreatedFiles()
    # We should be able to query created files even when none have been created
    assert created_files.by(module='name') == []

    a, b, c, d = create_temp_files(4)
    # Let us start by inserting one file and see if it is returned
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COPY,
        contents=[a],
        targets=[b],
    )
    assert created_files.by(module='name') == [b]

    # Insertion should be idempotent
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COPY,
        contents=[a],
        targets=[b],
    )
    assert created_files.by(module='name') == [b]

    # Now insert new files
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COPY,
        contents=[c],
        targets=[d],
    )
    assert created_files.by(module='name') == [b, d]

    # The content should be persisted across object lifetimes
    del created_files
    created_files = CreatedFiles()
    assert created_files.by(module='name') == [b, d]


def test_that_cleanup_method_removes_files(tmpdir, create_temp_files):
    """The cleanup method should remove all files created by module."""
    # Create some files which are made by modules
    (
        content1,
        content2,
        content3,
        content4,
        compilation1,
        compilation2,
        compilation4,
    ) = create_temp_files(7)
    symlink3 = Path(tmpdir, 'symlink3.tmp')
    symlink3.symlink_to(content3)

    # Tell CreatedFiles object that these files have been created by two
    # different modules
    created_files = CreatedFiles()
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COMPILE,
        contents=[content1, content2],
        targets=[compilation1, compilation2],
    )
    created_files.insert(
        module='name',
        creation_method=CreationMethod.SYMLINK,
        contents=[content3],
        targets=[symlink3],
    )
    created_files.insert(
        module='other_module',
        creation_method=CreationMethod.COMPILE,
        contents=[content4],
        targets=[compilation4],
    )

    # Cleanup files only made by name
    created_files.cleanup(module='name')

    # Content files should be left alone
    for content in (content1, content2, content3, content4):
        assert content.exists()

    # But target files should be removed
    for cleaned_file in (compilation1, compilation2, symlink3):
        assert not cleaned_file.exists()

    # The other module should be left alone
    assert compilation4.exists()

    # No files should be persisted as created by the module afterwards
    assert created_files.by(module='name') == []

    # This should be the case between object lifetimes too
    del created_files
    created_files = CreatedFiles()
    assert created_files.by(module='name') == []


def test_that_dry_run_is_respected(create_temp_files, caplog):
    """When dry_run is True, no files should be deleted."""
    content, target = create_temp_files(2)

    created_files = CreatedFiles()
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COPY,
        contents=[content],
        targets=[target],
    )

    caplog.clear()
    created_files.cleanup(module='name', dry_run=True)

    # None of the files should be affected
    assert content.exists()
    assert target.exists()

    # Skipping the deletion should be explicitly logged
    assert 'SKIPPED: ' in caplog.record_tuples[0][2]

    # And the files should still be considered created
    assert created_files.by(module='name') == [target]


def test_that_deleted_files_are_handled_gracefully_under_cleanup(
    create_temp_files,
):
    """When creations have been deleted they should be skipped."""
    content, target = create_temp_files(2)

    created_files = CreatedFiles()
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COPY,
        contents=[content],
        targets=[target],
    )

    # This should be gracefully handled
    target.unlink()
    created_files.cleanup(module='name')


def test_that_inserting_non_existent_file_is_skipped(
    create_temp_files,
):
    """When creations have been deleted they should be skipped."""
    content, target = create_temp_files(2)
    target.unlink()

    created_files = CreatedFiles()
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COPY,
        contents=[content],
        targets=[target],
    )

    # No file has been created!
    assert created_files.by(module='name') == []


def test_that_creations_are_properly_hashed(
    create_temp_files,
):
    """Creations should be hashed according to file content."""
    target1, target2, target3, content = create_temp_files(4)
    target1.write_text('identical')
    target2.write_text('identical')
    target3.write_text('different')

    created_files = CreatedFiles()
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COPY,
        contents=[content, content, content],
        targets=[target1, target2, target3],
    )

    # These hashes should be identical
    assert created_files.creations['name'][str(target1)]['hash'] \
        == created_files.creations['name'][str(target2)]['hash']

    # But these should be different
    assert created_files.creations['name'][str(target1)]['hash'] \
        != created_files.creations['name'][str(target3)]['hash']


def test_creating_created_files_object_for_specific_module(create_temp_files):
    """You should be able to construct a CreatedFiles wrapper for a module."""
    content, target = create_temp_files(2)
    created_files = CreatedFiles().wrapper_for(module='my_module')
    created_files.insert_creation(
        content=content,
        target=target,
        method=CreationMethod.SYMLINK,
    )


def test_backup_method(tmpdir, create_temp_files, patch_xdg_directory_standard):
    """Backup should perform backup of files not created by Astrality."""
    external_file, created_file, created_file_content = create_temp_files(3)
    external_file.write_text('original')
    created_file.write_text('new')
    created_file_content.write_text('content')

    created_files = CreatedFiles()

    # A module has created a file, without a collision
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COMPILE,
        contents=[created_file_content],
        targets=[created_file],
    )

    # The file should now be contained by the CreatedFiles object
    assert created_file in created_files

    # When we try to take a backup of the created file, we get None back,
    # as there is no need to perform a backup.
    assert created_files.backup(module='name', path=created_file) is None

    # The existing external file is *not* part of the created files
    assert external_file not in created_files

    # The creation should have no backup
    assert created_files.creations['name'][str(created_file)]['backup'] \
        is None

    # When we perform a backup, it is actually created
    backup = created_files.backup(module='name', path=external_file)

    # The backup contains the original filename in its new hashed filename
    assert external_file.name in backup.name

    # The content has been *moved* over
    assert backup.read_text() == 'original'
    assert not external_file.exists()

    # And the contents have been saved to the correct path
    assert backup.parent == patch_xdg_directory_standard / 'backups' / 'name'

    # The backup should now have been inserted
    assert created_files.creations['name'][str(external_file)]['backup'] \
        == str(backup)

    # If we now inform the creation, the backup info is kept
    shutil.copy2(str(created_file_content), str(external_file))
    created_files.insert(
        module='name',
        creation_method=CreationMethod.COPY,
        contents=[created_file_content],
        targets=[external_file],
    )
    assert created_files.creations['name'][str(external_file)]['backup']  \
        == str(backup)

    # But we also have the new information
    assert created_files.creations['name'][str(external_file)]['method']  \
        == 'copied'
    assert created_files.creations['name'][str(external_file)]['content'] \
        == str(created_file_content)

    # Before the cleanup, we have the new content in place
    assert external_file.read_text() == 'content'

    # But when we clean up the module, the backup should be put back in place
    created_files.cleanup('name')
    assert external_file.read_text() == 'original'


def test_taking_backup_of_symlinks(create_temp_files):
    """Symlinks should be properly backed up."""
    original_symlink, original_target, new_target = create_temp_files(3)

    # We have one existing symlink at the target location
    original_symlink.unlink()
    original_symlink.symlink_to(original_target)

    # The symlink points to original content
    original_target.write_text('original content')

    # Sanity check
    assert original_symlink.resolve() == original_target
    assert original_symlink.read_text() == 'original content'

    # We now backup the original symlink
    created_files = CreatedFiles()
    created_files.backup(module='name', path=original_symlink)

    # And we will now point it to new content
    new_target.write_text('new content')
    original_symlink.symlink_to(new_target)

    created_files.insert(
        module='name',
        creation_method=CreationMethod.SYMLINK,
        contents=[new_target],
        targets=[original_symlink],
    )

    # New sanity check
    assert original_symlink.resolve() == new_target
    assert original_symlink.read_text() == 'new content'

    # We now clean up the module and should get the original symlink back
    created_files.cleanup(module='name')
    assert original_symlink.resolve() == original_target
    assert original_symlink.read_text() == 'original content'


def test_creation_and_cleanup_of_directory(create_temp_files):
    """Created directories should be tracked."""
    # my_module is going to copy one file
    created_files = CreatedFiles().wrapper_for(module='my_module')

    # Specifically content -> target
    content, target = create_temp_files(2)

    # The target directory does not exist yet, so it creates it first
    created_files.insert_creation(
        content=None,
        target=target.parent,
        method=CreationMethod.MKDIR,
    )

    # Then copies the file over
    created_files.insert_creation(
        content=content,
        target=target,
        method=CreationMethod.COPY,
    )

    # These two creations should now be tracked
    global_created_files = CreatedFiles()
    creations = global_created_files.by('my_module')
    assert len(creations) == 2
    assert target.parent in global_created_files
    assert target in global_created_files

    # And a small sanity check, the directory actually exists
    assert target.parent.is_dir()

    # Now we introduce a small complication; an file **not** created by
    # astrality is placed within this created directory.
    external_file = target.parent / 'external.tmp'
    external_file.touch()

    # If we now clean up the module, the copied file can be deleted, but not
    # the created directory, as that would delete an external file!
    global_created_files.cleanup(module='my_module')
    assert not target.exists()
    assert target.parent.is_dir()

    # And the directory is still tracked, even after the cleanup
    assert target.parent in CreatedFiles()

    # Now we delet this external file,
    # such that the directory can be cleaned up.
    external_file.unlink()
    global_created_files.cleanup(module='my_module')
    assert not target.parent.is_dir()


def test_cleanup_of_recursive_directories(tmpdir):
    """Recursively created directories should be cleaned up."""
    tmpdir = Path(tmpdir)

    # Three recursive directories are created
    a = tmpdir / 'a'
    b = a / 'b'
    c = b / 'c'
    c.mkdir(parents=True)

    # And these creations are persisted
    created_files = CreatedFiles().wrapper_for(module='my_module')
    for directory in (c, a, b):
        created_files.insert_creation(
            content=None,
            target=directory,
            method=CreationMethod.MKDIR,
        )

    # And a file is copied over into the deepest of these directories
    content = tmpdir / 'content.tmp'
    target = c / 'target.tmp'
    content.touch()
    target.touch()
    created_files.insert_creation(
        content=content,
        target=target,
        method=CreationMethod.COPY,
    )

    # All these directories should be cleaned up
    CreatedFiles().cleanup(module='my_module')
    for directory in (a, b, c):
        assert not directory.exists()

    # Also the copied file
    assert not target.exists()


@pytest.mark.parametrize('with_wrapper', (False, True))
def test_mkdir_method_of_created_files(tmpdir, with_wrapper):
    """CreatedFiles should be able to create and persist directories."""
    tmpdir = Path(tmpdir)

    # Directory "a" already exists as a directory
    a = tmpdir / 'a'
    a.mkdir(parents=True)

    # But directory b and c are supposed to be created
    b = a / 'b'
    c = b / 'c'

    # We now want to create directory c and all parents
    created_files = CreatedFiles()

    if with_wrapper:
        created_files.wrapper_for(module='my_module').mkdir(path=c)
    else:
        created_files.mkdir(module='my_module', path=c)

    # The directories have been created
    assert b.is_dir()
    assert c.is_dir()

    # And persisted
    assert len(created_files.by('my_module')) == 2
    assert b in created_files
    assert c in created_files

    # But a should not be counted as a created directory
    assert a not in created_files

    # Cleanup should only clean b and c, but not a
    created_files.cleanup(module='my_module')
    assert not b.is_dir()
    assert not c.is_dir()
    assert a.is_dir()



"""Application wide fixtures."""
import os
from pathlib import Path
import shutil

import pytest

import astrality
from astrality.actions import ActionBlock
from astrality.config import GlobalModulesConfig, user_configuration
from astrality.context import Context
from astrality.module import Module, ModuleManager


@pytest.fixture
def conf_path():
    """Return str path to configuration directory."""
    conf_path = Path(__file__).parents[1] / 'config'
    return conf_path


@pytest.fixture
def conf_file_path(conf_path):
    """Return path to example configuration."""
    return conf_path / 'astrality.yml'


@pytest.fixture(scope='session', autouse=True)
def conf():
    """Return the configuration object for the example configuration."""
    this_test_file = os.path.abspath(__file__)
    conf_path = Path(this_test_file).parents[1] / 'config'
    return user_configuration(conf_path)[0]


@pytest.fixture(scope='session', autouse=True)
def context():
    """Return the context object for the example configuration."""
    this_test_file = os.path.abspath(__file__)
    conf_path = Path(this_test_file).parents[1] / 'config'
    return user_configuration(conf_path)[2]


@pytest.fixture(scope='session', autouse=True)
def modules():
    """Return the modules object for the example configuration."""
    this_test_file = os.path.abspath(__file__)
    conf_path = Path(this_test_file).parents[1] / 'config'
    return user_configuration(conf_path)[1]


@pytest.fixture
def test_config_directory():
    """Return path to test config directory."""
    return Path(__file__).parent / 'test_config'


@pytest.yield_fixture
def temp_directory(tmpdir):
    """Return path to temporary directory, and cleanup afterwards."""
    return Path(tmpdir).resolve()


@pytest.fixture
def context_directory(test_config_directory):
    """Return path to directory containing several context files."""
    return test_config_directory / 'context'


@pytest.fixture
def template_directory(test_config_directory):
    """Return path to directory containing several templates"""
    return test_config_directory / 'templates'


@pytest.fixture
def module_factory(test_config_directory):
    """Return Module factory for testing."""
    def _module_factory(
        name='test',
        on_startup=None,
        on_modified=None,
        on_exit=None,
        path=None,
        module_directory=test_config_directory / 'test_modules' /
        'using_all_actions',
        replacer=lambda x: x,
        context_store={},
    ) -> Module:
        """Return module with specified action blocks and config."""
        module = Module(
            name=name,
            module_config={},
            module_directory=module_directory,
            replacer=replacer,
            context_store=context_store,
        )
        if on_startup:
            module.action_blocks['on_startup'] = on_startup

        if on_exit:
            module.action_blocks['on_exit'] = on_exit

        if on_modified:
            module.action_blocks['on_modified'][path] = on_modified

        return module

    return _module_factory


@pytest.fixture
def module_manager_factory():
    """Return ModuleManager factory for testing."""
    def _module_manager_factory(
        *modules,
        context=Context(),
    ) -> ModuleManager:
        """Return ModuleManager object with given modules and context."""
        module_manager = ModuleManager(
            context=context,
        )
        module_manager.modules = {
            module.name: module
            for module
            in modules
        }

        # Insert correct context for all actions
        for module in modules:
            for block in module.all_action_blocks():
                for action_type in ActionBlock.action_types:
                    for actions in getattr(block, f'_{action_type}_actions'):
                        actions.context_store = context

        return module_manager

    return _module_manager_factory


@pytest.fixture
def create_temp_files(tmpdir):
    """Return temp file factory function."""
    temp_dir = Path(tmpdir)

    def _create_temp_files(number):
        """Create `number` tempfiles in seperate directories and yield paths."""
        for _number in range(number):
            temp_file = temp_dir / str(_number) / f'file{_number}.temp'
            temp_file.parent.mkdir(parents=True)
            temp_file.touch()
            yield temp_file

    return _create_temp_files


@pytest.fixture
def action_block_factory(test_config_directory):
    """Return action block factory function for testing."""

    def _action_block_factory(
        import_context={},
        compile={},
        copy={},
        run={},
        stow={},
        symlink={},
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store=Context(),
    ):
        """Return module with given parameters."""
        config = {
            'import_context': import_context,
            'compile': compile,
            'copy': copy,
            'run': run,
            'stow': stow,
            'symlink': symlink,
        }

        return ActionBlock(
            action_block=config,
            directory=directory,
            replacer=replacer,
            context_store=context_store,
            global_modules_config=GlobalModulesConfig(
                config={},
                config_directory=test_config_directory,
            ),
            module_name='test',
        )

    return _action_block_factory


@pytest.yield_fixture(autouse=True)
def patch_xdg_directory_standard(tmpdir, monkeypatch, request):
    """During testing, the XDG directory standard is monkeypatched."""
    if 'dont_patch_xdg' in request.keywords:
        yield
        return

    data_dir = Path(tmpdir).parent / '.local' / 'share' / 'astrality'

    # Clear data directory before the test
    if data_dir.exists():
        shutil.rmtree(str(data_dir))
    data_dir.mkdir(parents=True)

    monkeypatch.setattr(
        astrality.xdg.XDG,
        'data_home',
        data_dir,
    )

    yield data_dir

    # Delete directory for next test
    if data_dir.exists():
        shutil.rmtree(str(data_dir))


@pytest.fixture(autouse=True)
def patch_astrality_config_home(monkeypatch):
    """Patch $ASTRALITY_CONFIG_HOME."""
    example_config = Path(__file__).parents[2] / 'config'
    monkeypatch.setitem(
        os.environ,
        'ASTRALITY_CONFIG_HOME',
        str(example_config),
    )



"""Tests for finishing ModuleManager tasks with dry_run set to True."""

from astrality.module import ModuleManager


def test_that_dry_run_is_respected(create_temp_files):
    """ModuleManager should pass on dry_run to its actions."""
    (
        touched,
        copy_content,
        copy_target,
        compile_content,
        compile_target,
        symlink_content,
        symlink_target,
    ) = create_temp_files(7)
    touched.unlink()
    copy_target.write_text('copy_original')
    compile_target.write_text('compile_original')

    modules = {
        'A': {
            'run': {
                'shell': 'touch ' + str(touched),
            },
            'copy': {
                'content': str(copy_content),
                'target': str(copy_target),
            },
            'compile': {
                'content': str(compile_content), 'target': str(compile_target),
            },
            'symlink': {
                'content': str(symlink_content), 'target': str(symlink_target),
            },
        },
    }
    module_manager = ModuleManager(
        modules=modules,
        directory=touched.parents[1],
        dry_run=True,
    )
    module_manager.finish_tasks()

    assert not touched.exists()
    assert copy_target.read_text() == 'copy_original'
    assert compile_target.read_text() == 'compile_original'
    assert not symlink_target.is_symlink()



"""Tests for handling setup action blocks in ModuleManager."""

from pathlib import Path

from astrality.module import ModuleManager


def test_that_setup_block_is_only_executed_once(tmpdir):
    """Setup blocks in modules should only be performed once."""
    touched = Path(tmpdir, 'touched.tmp')
    modules = {
        'A': {
            'on_setup': {
                'run': {
                    'shell': f'touch {touched}',
                },
            },
        },
    }
    module_manager = ModuleManager(modules=modules)

    # The touched file should not exist before we have done anything
    assert not touched.exists()

    # After finishing tasks, the file should be touched
    module_manager.finish_tasks()
    assert touched.exists()

    # We now create a new object lifetime
    del module_manager
    touched.unlink()

    # The setup block should now *not* be executed
    module_manager = ModuleManager(modules=modules)
    module_manager.finish_tasks()
    assert not touched.exists()



"""Module for compilation of templates."""

import logging
import os
import shutil
from functools import partial
from pathlib import Path
from typing import Any, Dict, Optional

from jinja2 import (
    Environment,
    FileSystemLoader,
    Undefined,
    make_logging_undefined,
)

from astrality import utils
from astrality.context import Context

ApplicationConfig = Dict[str, Dict[str, Any]]
logger = logging.getLogger(__name__)


def jinja_environment(
    templates_folder: Path,
    shell_command_working_directory: Path,
) -> Environment:
    """Return a jinja Environment instance for templates in a folder."""
    logger = logging.getLogger(__name__)
    LoggingUndefined = make_logging_undefined(
        logger=logger,
        base=Undefined,
    )

    env = Environment(
        loader=FileSystemLoader(
            str(templates_folder),
            followlinks=True,
        ),
        autoescape=False,
        auto_reload=True,
        optimized=True,
        finalize=finalize_variable_expression,
        undefined=LoggingUndefined,
    )

    # Add env context containing all environment variables
    env.globals['env'] = os.environ

    # Add run shell command filter
    run_shell_from_working_directory = partial(
        utils.run_shell,
        working_directory=shell_command_working_directory,
        log_success=False,
    )
    env.filters['shell'] = run_shell_from_working_directory

    return env


def finalize_variable_expression(result: str) -> str:
    """Return empty strings for undefined template variables."""
    if result is None:
        return ''
    else:
        return result


def compile_template_to_string(
    template: Path,
    context: Context,
    shell_command_working_directory: Optional[Path] = None,
) -> str:
    """
    Return the compiled template string.

    Context placeholder replacements given by `context`, and shell filters
    run with working directory ``shell_command_working_directory``.
    """
    if not shell_command_working_directory:
        shell_command_working_directory = template.parent

    env = jinja_environment(
        templates_folder=template.parent,
        shell_command_working_directory=shell_command_working_directory,
    )
    jinja_template = env.get_template(name=template.name)

    return jinja_template.render(context)


def compile_template(
    template: Path,
    target: Path,
    context: Context,
    shell_command_working_directory: Path,
    permissions: Optional[str] = None,
) -> None:
    """
    Compile template to target destination with specific context.

    If `permissions` is provided, the target file will have its file mode set
    accordingly.
    permissions='755' -> chmod 755
    permissions='u+x' -> chmod u+x
    """
    logger.info(f'[Compiling] Template: "{template}" -> Target: "{target}"')

    result = compile_template_to_string(
        template=template,
        context=context,
        shell_command_working_directory=shell_command_working_directory,
    )

    # Create parent directories if they do not exist
    os.makedirs(target.parent, exist_ok=True)

    with open(target, 'w') as target_file:
        target_file.write(result)

    # Copy template's file permissions to compiled target file
    shutil.copymode(template, target)

    if permissions:
        result = utils.run_shell(
            command=f'chmod {permissions} {target}',
            timeout=0,
            fallback=False,
        )

        if result is False:
            logger.error(
                f'Could not set "{permissions}" permissions for "{target}"',
            )



"""Tests for astrality.persistence.ExecutedActions."""

import logging

from astrality.persistence import ExecutedActions


def test_that_executed_action_path_is_monkeypatched_in_all_files():
    """Autouse fixture should monkeypatch path property of ExecutedActions."""
    executed_actions = ExecutedActions(module_name='github::user/repo::module')
    path = executed_actions.path
    assert path.is_file()
    assert path.name == 'setup.yml'
    assert path.parent.name == 'astrality'
    assert 'test' in path.parents[3].name


def test_that_actions_are_saved_to_file(caplog):
    """Saved actions should be persisted properly."""
    action_option = {'shell': 'echo setup'}
    executed_actions = ExecutedActions(module_name='github::user/repo::module')

    # This action has not been executed yet.
    assert executed_actions.is_new(
        action_type='run',
        action_options=action_option,
    )

    # We save checked actions, and now it should count as executed.
    executed_actions.write()
    assert not executed_actions.is_new(
        action_type='run',
        action_options=action_option,
    )

    # The action should still be counted as executed between object lifetimes.
    del executed_actions
    executed_actions = ExecutedActions(module_name='github::user/repo::module')
    assert not executed_actions.is_new(
        action_type='run',
        action_options=action_option,
    )

    # But when we reset the module, it should be counted as new
    del executed_actions
    executed_actions = ExecutedActions(module_name='github::user/repo::module')

    caplog.clear()
    executed_actions.reset()
    assert executed_actions.is_new(
        action_type='run',
        action_options=action_option,
    )

    # Check that info is logged on success
    assert caplog.record_tuples[0][1] == logging.INFO

    # It should also be reset after loading from disk again
    executed_actions.reset()
    del executed_actions
    executed_actions = ExecutedActions(module_name='github::user/repo::module')
    assert executed_actions.is_new(
        action_type='run',
        action_options=action_option,
    )

    # Check that error is logged on invalid module name
    caplog.clear()
    ExecutedActions(module_name='i_do_not_exist').reset()
    assert caplog.record_tuples[0][1] == logging.ERROR



"""Module for test utilities."""

import re
import time
from typing import Any, Callable, Union


class RegexCompare:
    """
    Class for creating regex objects which can be compared with strings.

    :param regex: Regex string pattern"
    """
    def __init__(self, regex: str) -> None:
        """Constructor for RegexCompare object."""
        self.pattern = re.compile(regex)

    def __eq__(self, other) -> bool:
        """Return True if full regex match."""
        # Object only comparable with string
        assert isinstance(other, str)
        return bool(self.pattern.fullmatch(other))

    def __repr__(self) -> str:
        """Return string representation of RegexCompare object."""
        return f'RegexCompare({self.pattern})'


class Retry:
    """
    Class for retrying tests.

    :param expression: Zero arity callable which should return True after some
        retries.
    :param tries: Number of attempts.
    :param delay: Seconds to sleep between each attempt.
    :param increase_delay: Increase delay for each attempt.
    """

    def __init__(
        self,
        tries: int = 10,
        delay: Union[int, float] = 0.1,
        increase_delay: Union[int, float] = 0.3,
    ) -> None:
        """Retry object constructor."""
        self.tries = tries
        self.delay = delay
        self.increase = increase_delay

    def __call__(self, expression: Callable[[], Any]) -> bool:
        """
        Retry callable until it returns a truthy value.

        :param expression: Zero arity callable returning truhty/falsy values.
        """
        attempt = 0
        while attempt < self.tries:
            attempt += 1
            try:
                result = expression()
                if result:
                    return result
            except BaseException:
                pass

            time.sleep(self.delay)
            self.delay += self.increase

        return expression()



"""Tests for Context class."""
from math import inf
from pathlib import Path
import shutil

import pytest

from astrality.context import Context


class TestContextClass:
    def test_initialization_of_config_class_with_no_config_parser(self):
        Context()

    def test_invocation_of_class_with_application_config(self, conf):
        Context(conf)

    def test_initialization_of_config_class_with_dict(self):
        conf_dict = {
            'key1': 'value1',
            'key2': 'value2',
            'key3': ('one', 'two', 'three'),
            'key4': {'key4-1': 'uno', 'key4-2': 'dos'},
        }
        config = Context(conf_dict)
        assert config == conf_dict

    def test_values_for_max_key_property(self):
        config = Context()
        assert config._max_key == -inf

        config['string_key'] = 1
        assert config._max_key == -inf

        config[2] = 'string_value'
        assert config._max_key == 2

        config[1] = 'string_value'
        assert config._max_key == 2

        config[3] = 'string_value'
        assert config._max_key == 3

    def test_getting_item_from_empty_config(self):
        config = Context()
        with pytest.raises(KeyError):
            config['empty_config_with_no_key']

    def test_accessing_existing_key(self):
        config = Context()
        config['some_key'] = 'some_value'
        assert config['some_key'] == 'some_value'

        config[-2] = 'some_other_value'
        assert config[-2] == 'some_other_value'

    def test_integer_index_resolution(self):
        config = Context()
        config['some_key'] = 'some_value'
        config[1] = 'FureCode Nerd Font'
        assert config[2] == 'FureCode Nerd Font'

    def test_integer_index_resolution_without_earlier_index_key(self):
        config = Context()
        config['some_key'] = 'some_value'
        with pytest.raises(KeyError) as exception:
            config[2]
        assert exception.value.args[0] == \
            'Integer index "2" is non-existent and ' \
            'had no lower index to be substituted for'

    def test_index_resolution_with_string_key(self):
        config = Context()
        config[2] = 'some_value'
        with pytest.raises(KeyError) as exception:
            config['test']
        assert exception.value.args[0] == 'test'

    def test_use_of_recursive_config_objects_created_by_dicts(self):
        conf_dict = {
            'key1': 'value1',
            1: 'value2',
            2: {1: 'some_value'},
            'key3': ('one', 'two', 'three'),
            'key4': {1: 'uno', 'key4-2': 'dos'},
        }
        config = Context(conf_dict)
        assert config == conf_dict
        assert config[3][2] == 'some_value'
        assert config[2] == {1: 'some_value'}
        assert config[3] == {1: 'some_value'}

        assert isinstance(config['key4'], Context)
        assert config['key4'] == {1: 'uno', 'key4-2': 'dos'}
        assert config['key4'][1] == 'uno'
        assert config['key4'][2] == 'uno'

    def test_getter(self):
        config = Context()
        assert config.get('from_empty_config') is None

        config['test'] = 'something'
        assert config.get('test') == 'something'
        assert config.get('test', '4') == 'something'

        assert config.get('non_existent_key') is None
        assert config.get('non_existent_key', '4') == '4'

    def test_items(self):
        config = Context()
        config['4'] = 'test'
        config['font'] = 'Comic Sans'
        config['5'] = '8'
        assert list(config.items()) == [
            ('4', 'test'),
            ('font', 'Comic Sans'),
            ('5', '8'),
        ]

    def test_keys(self):
        config = Context()
        config['4'] = 'test'
        config['font'] = 'Comic Sans'
        config['5'] = '8'
        assert list(config.keys()) == ['4', 'font', '5']

    def test_values(self):
        config = Context()
        config['4'] = 'test'
        config['font'] = 'Comic Sans'
        config['5'] = '8'
        assert list(config.values()) == ['test', 'Comic Sans', '8']

    def test_update(self):
        one_conf_dict = {
            'key1': 'value1',
            1: 'value2',
            2: {1: 'some_value'},
        }
        another_conf_dict = {
            'key3': ('one', 'two', 'three'),
            'key4': {1: 'uno', 'key4-2': 'dos'},
        }
        merged_conf_dicts = {
            'key1': 'value1',
            1: 'value2',
            2: {1: 'some_value'},
            'key3': ('one', 'two', 'three'),
            'key4': {1: 'uno', 'key4-2': 'dos'},
        }
        config = Context(one_conf_dict)
        config.update(another_conf_dict)
        assert config == merged_conf_dicts

    def test_context_class(self):
        context = Context()
        context[1] = 'firs_value'
        context[2] = 'second_value'
        context['string_key'] = 'string_value'

        assert context[1] == 'firs_value'
        assert context[2] == 'second_value'
        assert context[3] == 'second_value'
        assert context['string_key'] == 'string_value'

    def test_initializing_context_with_context(self):
        context1 = Context({'key1': 1})
        context2 = Context(context1)
        assert context1 == context2

    def test_updating_context_with_context(self):
        context1 = Context({'key1': 1})
        context2 = Context({'key2': 2})

        context1.update(context2)
        expected_result = Context({'key1': 1, 'key2': 2})
        assert context1 == expected_result

    def test_reverse_updating_context(self):
        """Reverse update should not overwrite existing keys."""
        context1 = Context({
            'key1': 1,
            'key2': 2,
        })
        context2 = Context({
            'key2': 0,
            'key3': 0,
        })

        context1.reverse_update(context2)
        expected_result = Context({'key1': 1, 'key2': 2, 'key3': 0})
        assert context1 == expected_result


def test_importing_context_from_compiled_yml_file():
    """YAML files should be importable into the context."""
    context = Context({'section1': {'key_one': 'value_one'}})
    assert context['section1']['key_one'] == 'value_one'

    test_config_file = Path(__file__).parent / 'test_config' / 'test.yml'
    context.import_context(
        from_path=test_config_file,
        from_section='section2',
        to_section='new_section',
    )

    assert context['section1']['key_one'] == 'value_one'
    assert context['new_section']['var3'] == 'value1'

    context.import_context(
        from_path=test_config_file,
        from_section='section3',
        to_section='section3',
    )
    assert context['section3']['env_variable'] == 'test_value, hello'

    context.import_context(
        from_path=test_config_file,
        from_section='section1',
        to_section='section1',
    )
    assert context['section1']['var2'] == 'value1/value2'
    assert 'key_one' not in context['section1']


def test_instantiating_context_object_with_path():
    """Paths should be read into the context object."""
    test_context = Path(__file__).parent / 'test_config' / 'test.yml'
    context = Context(test_context)
    assert context == Context({
        'section1': {
            'var1': 'value1',
            'var2': 'value1/value2',
        },
        'section2': {
            'var3': 'value1',
            'empty_string_var': '',
        },
        'section3': {
            'env_variable': 'test_value, hello',
        },
        'section4': {
            1: 'primary_value',
        },
    })


def test_instantiating_with_directory_path(tmpdir):
    """Path directory should load context.yml."""
    test_context = Path(__file__).parent / 'test_config' / 'test.yml'
    shutil.copy(str(test_context), str(Path(tmpdir, 'context.yml')))

    context = Context(Path(tmpdir))
    assert context == Context({
        'section1': {
            'var1': 'value1',
            'var2': 'value1/value2',
        },
        'section2': {
            'var3': 'value1',
            'empty_string_var': '',
        },
        'section3': {
            'env_variable': 'test_value, hello',
        },
        'section4': {
            1: 'primary_value',
        },
    })



"""Tests for copy Module method."""

from pathlib import Path


def test_copying_in_on_modified_block(
    action_block_factory,
    create_temp_files,
    module_factory,
):
    """Module should copy properly."""
    file1, file2, file3, file4 = create_temp_files(4)
    file2.write_text('original')
    file4.write_text('some other content')

    action_block = action_block_factory(
        copy=[
            {'content': str(file1), 'target': str(file2)},
            {'content': str(file3), 'target': str(file4)},
        ],
    )
    module = module_factory(on_modified=action_block, path=Path('/a/b/c'))
    module.execute(action='all', block='on_modified', path=Path('/a/b/c'))

    # Check if content has been copied
    assert file2.read_text() == file1.read_text()
    assert file4.read_text() == file3.read_text()



"""Tests for ActionBlock class."""

from pathlib import Path

from astrality.actions import ActionBlock
from astrality.context import Context


def test_null_object_pattern(global_modules_config):
    """An empty action block should have no behaviour."""
    action_block = ActionBlock(
        action_block={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store=Context(),
        global_modules_config=global_modules_config,
        module_name='test',
    )
    action_block.execute(default_timeout=1)


def test_executing_action_block_with_one_action(
    global_modules_config,
    test_config_directory,
    tmpdir,
):
    """Action block behaviour with only one action specified."""
    temp_dir = Path(tmpdir)
    touched = temp_dir / 'touched.tmp'

    action_block_dict = {
        'run': [{'shell': 'touch ' + str(touched)}],
    }

    action_block = ActionBlock(
        action_block=action_block_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store=Context(),
        global_modules_config=global_modules_config,
        module_name='test',
    )

    action_block.execute(default_timeout=1)
    assert touched.is_file()


def test_executing_several_action_blocks(
    test_config_directory,
    tmpdir,
    global_modules_config,
):
    """Invoking execute() should execute all actions."""
    temp_dir = Path(tmpdir)
    target = temp_dir / 'target.tmp'
    touched = temp_dir / 'touched.tmp'

    action_block_dict = {
        'import_context': {'from_path': 'context/mercedes.yml'},
        'compile': [{
            'content': 'templates/a_car.template',
            'target': str(target),
        }],
        'run': {'shell': 'touch ' + str(touched)},
        'trigger': {'block': 'on_startup'},
    }
    context_store = Context()

    action_block = ActionBlock(
        action_block=action_block_dict,
        directory=test_config_directory,
        replacer=lambda x: x,
        context_store=context_store,
        global_modules_config=global_modules_config,
        module_name='test',
    )

    action_block.execute(default_timeout=1)
    assert context_store == {'car': {'manufacturer': 'Mercedes'}}
    assert target.read_text() == 'My car is a Mercedes'
    assert touched.is_file()


def test_retrieving_triggers_from_action_block(global_modules_config):
    """All trigger instructions should be returned."""
    action_block_dict = {
        'trigger': [
            {'block': 'on_startup'},
            {'block': 'on_modified', 'path': 'test.template'},
        ],
    }
    action_block = ActionBlock(
        action_block=action_block_dict,
        directory=Path('/'),
        replacer=lambda x: x,
        context_store=Context(),
        global_modules_config=global_modules_config,
        module_name='test',
    )

    startup_trigger, on_modified_trigger = action_block.triggers()

    assert startup_trigger.block == 'on_startup'
    assert on_modified_trigger.block == 'on_modified'
    assert on_modified_trigger.specified_path == 'test.template'
    assert on_modified_trigger.relative_path == Path('test.template')
    assert on_modified_trigger.absolute_path == Path('/test.template')


def test_retrieving_triggers_from_action_block_without_triggers(
    global_modules_config,
):
    """Action block with no triggers should return empty tuple."""
    action_block = ActionBlock(
        action_block={},
        directory=Path('/'),
        replacer=lambda x: x,
        context_store=Context(),
        global_modules_config=global_modules_config,
        module_name='test',
    )

    assert action_block.triggers() == tuple()


def test_retrieving_all_compiled_templates(
    global_modules_config,
    template_directory,
    tmpdir,
):
    """All earlier compilations should be retrievable."""
    template1 = template_directory / 'empty.template'
    template2 = template_directory / 'no_context.template'

    temp_dir = Path(tmpdir)
    target1 = temp_dir / 'target1.tmp'
    target2 = temp_dir / 'target2.tmp'
    target3 = temp_dir / 'target3.tmp'

    action_block_dict = {
        'compile': [
            {'content': str(template1), 'target': str(target1)},
            {'content': str(template1), 'target': str(target2)},
            {'content': str(template2), 'target': str(target3)},
        ],
    }

    action_block = ActionBlock(
        action_block=action_block_dict,
        directory=template_directory,
        replacer=lambda x: x,
        context_store=Context(),
        global_modules_config=global_modules_config,
        module_name='test',
    )

    assert action_block.performed_compilations() == {}

    action_block.execute(default_timeout=1)
    assert action_block.performed_compilations() == {
        template1: {target1, target2},
        template2: {target3},
    }


def test_symlinking(action_block_factory, create_temp_files):
    """Action blocks should symlink properly."""
    file1, file2, file3, file4 = create_temp_files(4)
    file2.write_text('original')

    action_block = action_block_factory(
        symlink=[
            {'content': str(file1), 'target': str(file2)},
            {'content': str(file3), 'target': str(file4)},
        ],
    )
    action_block.symlink()

    assert file2.is_symlink()
    assert file2.resolve() == file1
    assert file4.is_symlink()
    assert file4.resolve() == file3


def test_copying(action_block_factory, create_temp_files):
    """Action blocks should copy properly."""
    file1, file2, file3, file4 = create_temp_files(4)
    file2.write_text('original')
    file4.write_text('some other content')

    action_block = action_block_factory(
        copy=[
            {'content': str(file1), 'target': str(file2)},
            {'content': str(file3), 'target': str(file4)},
        ],
    )
    action_block.copy()

    # Check if content has been copied
    assert file2.read_text() == file1.read_text()
    assert file4.read_text() == file3.read_text()


def test_stowing(action_block_factory, create_temp_files):
    """Action blocks should stow properly."""
    template, target = create_temp_files(2)
    template.write_text('{{ env.EXAMPLE_ENV_VARIABLE }}')
    symlink_target = template.parent / 'symlink_me'
    symlink_target.touch()

    action_block = action_block_factory(
        stow={
            'content': str(template.parent),
            'target': str(target.parent),
            'templates': r'file(0).temp',
            'non_templates': 'symlink',
        },
    )
    action_block.stow()

    # Check if template has been compiled
    assert Path(target.parent / '0').read_text() == 'test_value'

    # Check if non_template has been symlinked
    assert (template.parent / 'symlink_me').resolve() == symlink_target



"""Module implementing XDG directory standard for astrality."""

import os
from pathlib import Path


class XDG:
    """
    Class for handling the XDG directory standard.

    :param application_name: Name of application to use XDG directory standard.
    """

    def __init__(self, application_name: str = 'astrality') -> None:
        """Contstruct XDG object for application."""
        self.application_name = application_name
        self.XDG_DATA_HOME = Path(os.environ.get(
            'XDG_DATA_HOME',
            '~/.local/share',
        )).expanduser()
        self.XDG_CONFIG_HOME = Path(os.environ.get(
            'XDG_CONFIG_HOME',
            '~/.config',
        )).expanduser()

    @property
    def data_home(self) -> Path:
        """
        Return XDG_DATA_HOME directory of application.

        :return: Path to resolved value of XDG_DATA_HOME.
        """
        application_data_home = self.XDG_DATA_HOME / self.application_name
        application_data_home.mkdir(parents=True, exist_ok=True)
        return application_data_home

    def data(self, resource: str, directory: bool = False) -> Path:
        """
        Return file resource of application.

        :param resource: Relative string path, i.e. $XDG_DATA_HOME/<resource>.
        :param directory: If True, a directory is returned instead of a file.
        :return: Path to touched data resource.
        """
        resource_path = self.data_home / resource
        resource_path.parent.mkdir(parents=True, exist_ok=True)

        if directory:
            resource_path.mkdir(exist_ok=True)
        else:
            resource_path.touch(exist_ok=True)

        return resource_path



"""
Module defining class-representation of module actions.

Each action class type encapsulates the user specified options available for
that specific action type. The action itself can be performed by invoking the
object method `execute()`.

One of the main goals with Action, is that the arity of execute is 0.
This means that we unfortunately need to pass a reference to global mutable
state, i.e. the context store.

Another goal is that none of the subclasses require the global configuration
of the entire application, just the action configuration itself. Earlier
implementations required GlobalApplicationConfig to be passed arround in the
entire run-stack, which was quite combersome. Some of the limitations with this
approach could be solved if we implement GlobalApplicationConfig as a singleton
which could be imported and accessed independently from other modules.
"""

import abc
import hashlib
import logging
import os
from collections import defaultdict
from pathlib import Path
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from mypy_extensions import TypedDict

from astrality import compiler, utils
from astrality import config
from astrality.context import Context
from astrality import persistence
from astrality.xdg import XDG

Replacer = Callable[[str], str]


class Action(abc.ABC):
    """
    Superclass for module action types.

    :param options: A dictionary containing the user options for a given module
        action type.
    :param directory: The directory used as anchor for relative paths. This
        must be an absolute path.
    :param replacer: Placeholder substitutor of string user options.
    :param context_store: A reference to the global context store.
    :param creation_store: ModuleCreatedFiles object which stores which files
        that are created by the different module actions.
    """

    directory: Path
    priority: int
    Options = Union[
        'CompileDict',
        'CopyDict',
        'ImportContextDict',
        'RunDict',
        'StowDict',
        'SymlinkDict',
        'TriggerDict',
    ]

    def __init__(
        self,
        options: 'Action.Options',
        directory: Path,
        replacer: Replacer,
        context_store: Context,
        creation_store: 'persistence.ModuleCreatedFiles',
    ) -> None:
        """Contstruct action object."""
        # If no options are provided, use null object pattern
        self.null_object = not bool(options)

        assert directory.is_absolute()
        self.directory = directory
        self.context_store = context_store
        self.creation_store = creation_store
        self._options = options
        self._replace = replacer

    def replace(self, string: str) -> str:
        """
        Return converted string, substitution defined by `replacer`.

        This is used to replace placeholders such as {event}.
        This redirection is necessary due to python/mypy/issues/2427

        :param string: String configuration option.
        :return: String with placeholders substituted.
        """
        return self._replace(string)

    def option(self, key: str, default: Any = None, path: bool = False) -> Any:
        """
        Return user specified action option.

        All option value access should go through this helper function, as
        it replaces relevant placeholders users might have specified.

        :param key: The key of the user option that should be retrieved.
        :param default: Default return value if key not found.
        :param path: If True, convert string path to Path.is_absolute().
        :return: Processed action configuration value.
        """
        option_value = self._options.get(key, default)

        if option_value is None:
            return None
        elif path:
            # The option value represents a path, that should be converted
            # to an absolute pathlib.Path object
            assert isinstance(option_value, str)
            substituted_string_path = self.replace(option_value)
            return self._absolute_path(of=substituted_string_path)
        elif isinstance(option_value, str):
            # The option is a string, and any placeholders should be
            # substituted before it is returned. We also expand any environment
            # variables that might be present.
            return os.path.expandvars(self.replace(option_value))
        else:
            return option_value

    def _absolute_path(self, of: str) -> Path:
        """
        Return absolute path from relative string path.

        :param of: Relative path.
        :return: Absolute path anchored to `self.directory`.
        """
        return config.expand_path(
            path=Path(of),
            config_directory=self.directory,
        )

    @abc.abstractmethod
    def execute(self, dry_run: bool = False) -> Any:
        """
        Execute defined action.

        :param dry_run: If external side effects should be skipped.
        """

    def __repr__(self) -> str:
        """Return string representation of Action object."""
        return self.__class__.__name__ + f'({self._options})'


class RequiredImportContextDict(TypedDict):
    """Required fields of a import_context action."""

    from_path: str


class ImportContextDict(RequiredImportContextDict, total=False):
    """Allowable fields of an import_context action."""

    from_section: str
    to_section: str


class ImportContextAction(Action):
    """
    Import context into global context store.

    :param context_store: A mutable reference to the global context store.

    See :class:`Action` for documentation for the other parameters.
    """

    priority = 100
    context_store: compiler.Context

    def execute(self, dry_run: bool = False) -> None:
        """
        Import context section(s) according to user configuration block.

        :param dry_run: This parameter is ignored, as import_context only has
            internal side effects.
        """
        if self.null_object:
            # Null object does nothing
            return None

        self.context_store.import_context(
            from_path=self.option(key='from_path', path=True),
            from_section=self.option(key='from_section'),
            to_section=self.option(key='to_section'),
        )


class RequiredCompileDict(TypedDict):
    """Required fields of compile action."""

    content: str


class CompileDict(RequiredCompileDict, total=False):
    """Allowable fields of compile action."""

    target: str
    include: str
    permissions: str


class CompileAction(Action):
    """Compile template action."""

    _options: CompileDict

    priority = 400

    def __init__(self, *args, **kwargs) -> None:
        """Construct compile action object."""
        super().__init__(*args, **kwargs)
        self._performed_compilations: DefaultDict[Path, Set[Path]] = \
            defaultdict(set)

    def execute(self, dry_run: bool = False) -> Dict[Path, Path]:
        """
        Compile template source to target destination.

        :param dry_run: If True, skip and log compilation(s).
        :return: Dictionary with template content keys and target path values.
        """
        if self.null_object:
            # Null objects do nothing
            return {}
        elif 'target' not in self._options:
            # If no target is specified, we create a deterministic target.
            template = self.option(key='content', path=True)
            target = self.create_compilation_target(template=template)
            self._options['target'] = str(target)

        # These might either be file paths or directory paths
        template_source = self.option(key='content', path=True)
        target_source = self.option(key='target', path=True)
        if not template_source.exists():
            logger = logging.getLogger(__name__)
            logger.error(
                f'Could not compile template "{template_source}" '
                f'to target "{target_source}". No such path!',
            )
            return {}

        compile_pairs = utils.resolve_targets(
            content=template_source,
            target=target_source,
            include=self.option(key='include', default=r'(.+)'),
        )
        permissions = self.option(key='permissions')

        for content_file, target_file in compile_pairs.items():
            if dry_run:
                logger = logging.getLogger(__name__)
                logger.info(
                    f'SKIPPED: '
                    f'[Compiling] Template: "{content_file}" '
                    f'-> Target: "{target_file}"',
                )
            else:
                self.creation_store.backup(path=target_file)
                self.creation_store.mkdir(path=target_file.parent)
                compiler.compile_template(
                    template=content_file,
                    target=target_file,
                    context=self.context_store,
                    shell_command_working_directory=self.directory,
                    permissions=permissions,
                )
                self.creation_store.insert_creation(
                    content=content_file,
                    target=target_file,
                    method=persistence.CreationMethod.COMPILE,
                )

            self._performed_compilations[content_file].add(target_file)

        return compile_pairs

    def performed_compilations(self) -> DefaultDict[Path, Set[Path]]:
        """
        Return dictionary containing all performed compilations.

        :return: Dictinary with keys containing compiled templates, and values
            as a set of target paths.
        """
        return self._performed_compilations.copy()

    def create_compilation_target(self, template: Path) -> Path:
        """
        Create compilation target for template with unspecified target.

        Compilation targets are stored in $XDG_DATA_HOME/astrality/compilations.
        For details regarding the implementation see:
        https://www.peterbe.com/plog/best-hashing-function-in-python

        :param name: Path to template to be compiled.
        :return: Path to deterministicly determined compilation target.
        """
        # First dump the action configuration to YAML formatted string
        yaml_config = utils.yaml_str(self._options)

        # Now encode the string into raw bytes
        yaml_config_bytes = yaml_config.encode('utf-8', errors='ignore')

        # Create a MD5 hash from the string, only using the first seven chars
        yaml_config_md5 = hashlib.md5(yaml_config_bytes).hexdigest()[:7]

        # Prepend the template name for readability
        unique_name = template.name + '-' + yaml_config_md5

        # Create compilation target in XDG data directory
        compile_target = XDG().data(resource='compilations/' + unique_name)
        return compile_target

    def __contains__(self, other) -> bool:
        """Return True if run action is responsible for template."""
        assert other.is_absolute()

        if not self.option(key='content', path=True) == other:
            # This is not a managed template, so we will not recompile
            return False

        # Return True if the template has been compiled
        return other in self.performed_compilations()


class RequiredSymlinkDict(TypedDict):
    """Required fields of symlink action user config."""

    content: str
    target: str


class SymlinkDict(RequiredSymlinkDict, total=False):
    """Allowable fields of symlink action user config."""

    include: str


class SymlinkAction(Action):
    """Symlink files Action sub-class."""

    priority = 200

    _options: SymlinkDict

    def __init__(self, *args, **kwargs) -> None:
        """Construct symlink action object."""
        super().__init__(*args, **kwargs)
        self.symlinked_files: DefaultDict[Path, Set[Path]] = \
            defaultdict(set)

    def execute(self, dry_run: bool = False) -> Dict[Path, Path]:
        """
        Symlink to `content` path from `target` path.

        :param dry_run: If True, skip and log symlink creation(s).
        :return: Dictionary with content keys and symlink values.
        """
        if self.null_object:
            return {}

        content = self.option(key='content', path=True)
        target = self.option(key='target', path=True)
        include = self.option(key='include', default=r'(.+)')
        links = utils.resolve_targets(
            content=content,
            target=target,
            include=include,
        )

        logger = logging.getLogger(__name__)
        for content, symlink in links.items():
            self.symlinked_files[content].add(symlink)
            log_msg = f'[symlink] Content "{content}" -> Target: "{symlink}".'

            if symlink.resolve() == content.resolve():
                continue

            if dry_run:
                logger.info('SKIPPED: ' + log_msg)
                continue

            logger.info(log_msg)
            self.creation_store.mkdir(symlink.parent)
            self.creation_store.backup(path=symlink)
            symlink.symlink_to(content)
            self.creation_store.insert_creation(
                content=content,
                target=symlink,
                method=persistence.CreationMethod.SYMLINK,
            )

        return links


class RequiredCopyDict(TypedDict):
    """Required fields of copy action user config."""

    content: str
    target: str


class CopyDict(RequiredCopyDict, total=False):
    """Allowable fields of copy action user config."""

    include: str
    permissions: str


class CopyAction(Action):
    """Copy files Action sub-class."""

    priority = 300

    _options: CopyDict

    def __init__(self, *args, **kwargs) -> None:
        """Construct copy action object."""
        super().__init__(*args, **kwargs)
        self.copied_files: DefaultDict[Path, Set[Path]] = \
            defaultdict(set)

    def execute(self, dry_run: bool = False) -> Dict[Path, Path]:
        """
        Copy from `content` path to `target` path.

        :param dry_run: If True, skip and log copy creation(s).
        :return: Dictionary with content keys and copy values.
        """
        if self.null_object:
            return {}

        content = self.option(key='content', path=True)
        target = self.option(key='target', path=True)
        include = self.option(key='include', default=r'(.+)')
        permissions = self.option(key='permissions', default=None)

        copies = utils.resolve_targets(
            content=content,
            target=target,
            include=include,
        )
        logger = logging.getLogger(__name__)
        for content, copy in copies.items():
            self.copied_files[content].add(copy)

            log_msg = f'[copy] Content: "{content}" -> Target: "{target}".'
            if dry_run:
                logger.info('SKIPPED: ' + log_msg)
                continue

            logger.info(log_msg)
            self.creation_store.mkdir(copy.parent)
            self.creation_store.backup(path=copy)
            utils.copy(
                source=content,
                destination=copy,
                follow_symlinks=False,
            )
            self.creation_store.insert_creation(
                content=content,
                target=copy,
                method=persistence.CreationMethod.COPY,
            )

        if permissions and not dry_run:
            for copy in copies.values():
                result = utils.run_shell(
                    command=f'chmod {permissions} "{copy}"',
                    timeout=1,
                    fallback=False,
                )

                if result is False:
                    logger = logging.getLogger(__name__)
                    logger.error(
                        f'Could not set "{permissions}" '
                        f'permissions for copy "{target}"',
                    )

        return copies

    def __contains__(self, other) -> bool:
        """Return True if path has been copied *from*."""
        return other in self.copied_files


class RequiredStowDict(TypedDict):
    """Required dictionary keys for user stow action config."""

    content: str
    target: str


class StowDict(RequiredStowDict, total=False):
    """Allowable dictionary keys for user stow action config."""

    templates: str
    non_templates: str
    permissions: str


class StowAction(Action):
    """Stow directory action."""

    non_templates_action: Union[CopyAction, SymlinkAction]
    _options: StowDict

    priority = 500

    def __init__(self, *args, **kwargs) -> None:
        """Construct stow action object."""
        super().__init__(*args, **kwargs)
        if self.null_object:
            return

        # Create equivalent compile action based on stow config
        compile_options: CompileDict = {
            'content': self._options['content'],
            'target': self._options['target'],
            'include': self._options.get('templates', r'template\.(.+)'),
        }
        if 'permissions' in self._options:
            compile_options['permissions'] = self._options['permissions']

        self.compile_action = CompileAction(
            options=compile_options,
            directory=self.directory,
            replacer=self.replace,
            context_store=self.context_store,
            creation_store=self.creation_store,
        )

        # Determine what to do with non-templates
        non_templates_action = self._options.get('non_templates', 'symlink')
        self.ignore_non_templates = non_templates_action.lower() == 'ignore'

        if non_templates_action.lower() not in ('copy', 'symlink', 'ignore'):
            logger = logging.getLogger(__name__)
            logger.error(
                f'Invalid stow non_templates parameter:'
                f'"{non_templates_action}". '
                'Should be one of "symlink", "copy", or "ignore"!',
            )
            self.ignore_non_templates = True
            return

        # Negate the `templates` regex pattern in order to match non-templates
        if 'templates' in self._options:
            excluded = r'(?!' + self._options['templates'] + r').+'
        else:
            excluded = r'(?!template\..+).+'

        # Create configuration used for either symlink or copy
        non_templates_options: Dict = {
            'content': self._options['content'],
            'target': self._options['target'],
            'include': excluded,
            'permissions': self._options.get('permissions'),
        }

        # Create action object based on parameter `non_templates`
        NonTemplatesAction: Union[Type[CopyAction], Type[SymlinkAction]]
        if non_templates_action.lower() == 'copy':
            NonTemplatesAction = CopyAction
        else:
            NonTemplatesAction = SymlinkAction

        self.non_templates_action = NonTemplatesAction(
            options=non_templates_options,
            directory=self.directory,
            replacer=self.replace,
            context_store=self.context_store,
            creation_store=self.creation_store,
        )

    def execute(self, dry_run: bool = False) -> Dict[Path, Path]:
        """
        Stow directory source to target destination.

        :param dry_run: If True, skip and log copies, symlinks, and
            compilations.
        :return: Dictionary with source keys and target values.
            Contains compiled, symlinked, and copied files.
        """
        if self.null_object:
            return {}

        if self.ignore_non_templates:
            return self.compile_action.execute(dry_run=dry_run)
        else:
            copies_or_links = self.non_templates_action.execute(dry_run=dry_run)
            compilations = self.compile_action.execute(dry_run=dry_run)
            compilations.update(copies_or_links)
            return compilations

    def managed_files(self) -> Dict[Path, Set[Path]]:
        """
        Return dictionary containing content keys and target values.

        :return: Dictinary with keys containing compiled templates, and values
            as a set of target paths. If `non_templates` is 'copy', then these
            will be included as well.
        """
        if self.null_object:
            return {}

        managed_files = self.compile_action._performed_compilations.copy()

        if isinstance(self.non_templates_action, CopyAction):
            managed_files.update(self.non_templates_action.copied_files)

        return managed_files

    def __contains__(self, other) -> bool:
        """
        Return True if stow action is responsible for file path.

        A stow action is considered to be responsible for a file path if that
        path is modified results in its tasks to be outdated, and it needs to be
        re-executed.

        :param other: File path.
        :return: Boolean indicating if path has been copied or compiled.
        """
        assert other.is_absolute()
        return other in self.managed_files()


class RunDict(TypedDict):
    """Required fields of run action user config."""

    shell: str
    timeout: Union[int, float]


class RunAction(Action):
    """Run shell command Action sub-class."""

    _options: RunDict

    priority = 600

    def execute(  # type: ignore
        self,
        default_timeout: Union[int, float] = 0,
        dry_run: bool = False,
    ) -> Optional[Tuple[str, str]]:
        """
        Execute shell command action.

        :param default_timeout: Run timeout in seconds if no specific value is
            specified in `options`.
        :param dry_run: If True, skip and log commands to be executed.
        :return: 2-tuple containing the executed command and its resulting
            stdout.
        """
        if self.null_object:
            # Null objects do nothing
            return None

        command = self.option(key='shell')
        timeout = self.option(key='timeout')

        logger = logging.getLogger(__name__)
        if dry_run:
            logger.info(
                f'SKIPPED: [run] Command: "{command}" (timeout={timeout}).',
            )
            return command, ''

        logger.info(f'Running command "{command}".')
        result = utils.run_shell(
            command=command,
            timeout=timeout or default_timeout,
            working_directory=self.directory,
        )
        return command, result


class TriggerDictRequired(TypedDict):
    """Required fields of a trigger module action."""

    block: str


class TriggerDict(TriggerDictRequired, total=False):
    """Optional fields of a trigger module action."""

    path: str


class Trigger:
    """
    A class representing an instruction to trigger a specific action block.

    :ivar block: The block to be trigger, for example 'on_startup',
        'on_event', 'on_exit', or 'on_modified'.
    :ivar specified_path: The string path specified for a 'on_modified' block.
    :ivar relative_path: The relative pathlib.Path specified by
        `specified_path`.
    :ivar absolute_path: The absolute path specified by `specified_path`.
    """

    block: str
    specified_path: Optional[str]
    relative_path: Optional[Path]
    absolute_path: Optional[Path]

    def __init__(
        self,
        block: str,
        specified_path: Optional[str] = None,
        relative_path: Optional[Path] = None,
        absolute_path: Optional[Path] = None,
    ) -> None:
        """Construct trigger instruction."""
        self.block = block
        self.specified_path = specified_path
        self.relative_path = relative_path
        self.absolute_path = absolute_path


class TriggerAction(Action):
    """Action sub-class representing a trigger action."""

    _options: TriggerDict

    priority = 0

    def execute(self, dry_run: bool = False) -> Optional[Trigger]:
        """
        Return trigger instruction.

        If no trigger is specified, return None.

        :param dry_run: This parameter is ignored.
        :return: Optional :class:`.Trigger` instance.
        """
        if self.null_object:
            """Null objects do nothing."""
            return None

        block = self.option(key='block')

        if block != 'on_modified':
            # We do not need any paths, as the trigger block is not relative to
            # any modified path.
            return Trigger(block=block)

        # The modified path specified by the user configuration
        specified_path = self.option(key='path')

        # Instantiate relative and absolute pathlib.Path objects
        relative_path = Path(specified_path)
        absolute_path = self._absolute_path(of=specified_path)

        # Return 'on_modified' Trigger object with path information
        return Trigger(
            block=block,
            specified_path=specified_path,
            relative_path=relative_path,
            absolute_path=absolute_path,
        )


class ActionBlockDict(TypedDict, total=False):
    """Valid keys in an action block."""

    import_context: Union[ImportContextDict, List[ImportContextDict]]
    compile: Union[CompileDict, List[CompileDict]]
    run: Union[RunDict, List[RunDict]]
    trigger: Union[TriggerDict, List[TriggerDict]]


class ActionBlock:
    """
    Class representing a module action block, e.g. 'on_startup'.

    :param action_block: Dictionary containing all actions to be performed.
    :param directory: The directory used as anchor for relative paths. This
        must be an absolute path.
    :param replacer: Placeholder substitutor of string user options.
    :param context_store: A reference to the global context store.
    :param module_name: Name of module owning ActionBlock.
    :param global_modules_config: Global configuration object.
    """

    _compile_actions: List[CompileAction]
    _copy_actions: List[CopyAction]
    _import_context_actions: List[ImportContextAction]
    _run_actions: List[RunAction]
    _stow_actions: List[StowAction]
    _symlink_actions: List[SymlinkAction]
    _trigger_actions: List[TriggerAction]

    action_types = {
        'import_context': ImportContextAction,
        'symlink': SymlinkAction,
        'copy': CopyAction,
        'compile': CompileAction,
        'stow': StowAction,
        'run': RunAction,
        'trigger': TriggerAction,
    }

    def __init__(
        self,
        action_block: ActionBlockDict,
        directory: Path,
        replacer: Replacer,
        context_store: compiler.Context,
        global_modules_config: 'config.GlobalModulesConfig',
        module_name: str,
    ) -> None:
        """Construct ActionBlock object."""
        assert directory.is_absolute()
        if not global_modules_config:
            # TODO: This is only the case in testing, and should be explicitly
            # provided in all tests at some point.
            global_modules_config = config.GlobalModulesConfig(
                config={},
                config_directory=directory,
            )
        self.action_block = action_block
        self.module_name = module_name
        self.run_timeout = global_modules_config.run_timeout

        creation_store = global_modules_config.created_files.wrapper_for(
            module=self.module_name,
        )
        for identifier, action_type in self.action_types.items():
            # Create and persist a list of all ImportContextAction objects
            setattr(
                self,
                f'_{identifier}_actions',
                [
                    action_type(
                        options=action_options,
                        directory=directory,
                        replacer=replacer,
                        context_store=context_store,
                        creation_store=creation_store,
                    )
                    for action_options
                    in self.action_options(identifier=identifier)
                ],
            )

    def action_options(self, identifier: str) -> List[Action.Options]:
        """
        Return all action configs of type 'identifier'.

        :param identifier: Action type, such as 'run' or 'compile'.
        :return: List of action options of that type.
        """
        return utils.cast_to_list(                  # type: ignore
            self.action_block.get(identifier, {}),  # type: ignore
        )

    def import_context(self, dry_run: bool = False) -> None:
        """Import context into global context store."""
        for import_context_action in self._import_context_actions:
            import_context_action.execute(dry_run=dry_run)

    def symlink(self, dry_run: bool = False) -> None:
        """Symlink files."""
        for symlink_action in self._symlink_actions:
            symlink_action.execute(dry_run=dry_run)

    def copy(self, dry_run: bool = False) -> None:
        """Copy files."""
        for copy_action in self._copy_actions:
            copy_action.execute(dry_run=dry_run)

    def compile(self, dry_run: bool = False) -> None:
        """Compile templates."""
        for compile_action in self._compile_actions:
            compile_action.execute(dry_run=dry_run)

    def stow(self, dry_run: bool = False) -> None:
        """Stow directory contents."""
        for stow_action in self._stow_actions:
            stow_action.execute(dry_run=dry_run)

    def run(
        self,
        default_timeout: Optional[Union[int, float]] = None,
        dry_run: bool = False,
    ) -> Tuple[Tuple[str, str], ...]:
        """
        Run shell commands.

        :param default_timeout: How long to wait for run commands to exit
        :return: Tuple of 2-tuples containing (shell_command, stdout,)
        """
        results: Tuple[Tuple[str, str], ...] = tuple()
        for run_action in self._run_actions:
            result = run_action.execute(
                default_timeout=default_timeout or self.run_timeout,
                dry_run=dry_run,
            )
            if result:
                # Run action is not null object, so we can return results
                command, stdout = result
                results += ((command, stdout),)

        return results

    def triggers(self, dry_run: bool = False) -> Tuple[Trigger, ...]:
        """
        Return all trigger instructions specified in action block.

        :return: Tuple of Trigger objects specified in action block.
        """
        return tuple(
            trigger_action.execute(dry_run=dry_run)  # type: ignore
            for trigger_action
            in self._trigger_actions
            if not trigger_action.null_object
        )

    def execute(
        self,
        default_timeout: Union[int, float],
        dry_run: bool = False,
    ) -> None:
        """
        Execute all actions in action block.

        The order of execution is:
            1) Perform all context imports into the context store.
            2) Compile all templates.
            3) Run all shell commands.
        """
        self.import_context()
        self.compile()
        self.run(default_timeout=default_timeout, dry_run=dry_run)

    def performed_compilations(self) -> DefaultDict[Path, Set[Path]]:
        """
        Return all earlier performed compilations.

        :return: Dictionary with template keys and target path set.
        """
        all_compilations: DefaultDict[Path, Set[Path]] = defaultdict(set)
        for compile_action in self._compile_actions:
            compilations = compile_action.performed_compilations()
            for template, targets in compilations.items():
                all_compilations[template] |= targets

        return all_compilations


class SetupActionBlock(ActionBlock):
    """Setup action block which only executes actions once."""

    def action_options(self, identifier: str) -> List[Action.Options]:
        """
        Return action configs of 'identifier' type that have not been executed.

        :param identifier: Action type, such as 'run' or 'compile'.
        :return: List of action options of that type.
        """
        action_options = super().action_options(identifier)

        if not hasattr(self, 'executed_setup_actions'):
            self.executed_setup_actions = persistence.ExecutedActions(
                module_name=self.module_name,
            )

        not_executed = [
            action_option
            for action_option
            in action_options
            if self.executed_setup_actions.is_new(
                action_type=identifier,
                action_options=action_option,
            )
        ]
        self.executed_setup_actions.write()
        return not_executed



import os
import shutil
from sys import platform
from pathlib import Path

import pytest
from watchdog.observers import Observer

from astrality.filewatcher import DirectoryWatcher
from astrality.tests.utils import Retry


MACOS = platform == 'darwin'


@pytest.yield_fixture
def watch_dir(tmpdir):
    """Instanciate a directory watcher and stop it after its use."""
    watched_directory = Path(tmpdir)
    test_file1 = watched_directory / 'tmp_test_file1'
    recursive_dir = watched_directory / 'test_folder'
    test_file2 = recursive_dir / 'tmp_test_file2'

    class EventSaver:
        """Mock class for testing callback function."""

        def __init__(self):
            self.called = 0
            self.argument = None

        def save_argument(self, path: Path) -> None:
            self.called += 1
            self.argument = path

    event_saver = EventSaver()

    # Watch a temporary directory
    dir_watcher = DirectoryWatcher(
        directory=watched_directory,
        on_modified=event_saver.save_argument,
    )

    yield (
        watched_directory,
        recursive_dir,
        test_file1,
        test_file2,
        dir_watcher,
        event_saver,
    )

    dir_watcher.stop()

    # Cleanup files
    if test_file1.is_file():
        os.remove(test_file1)
    if test_file2.is_file():
        os.remove(test_file2)
    if recursive_dir.is_dir():
        shutil.rmtree(recursive_dir)


@pytest.mark.skipif(MACOS, reason='Flaky on MacOS')
@pytest.mark.slow
def test_filesystem_watcher(watch_dir):
    """
    Test correct callback invocation on directory watching.

    Sometimes the on_modified function is called several times by watchdog,
    for a unknown reason. It might be other tests which interfer. We therefore
    check if the lower bound of calls is satisfied, but do not test the exact
    number of calls to on_modified.
    """
    (
        watched_directory,
        recursive_dir,
        test_file1,
        test_file2,
        dir_watcher,
        event_saver,
    ) = watch_dir

    # Start watching the directory
    dir_watcher.start()

    # Nothing has been modified yet
    assert event_saver.argument is None
    assert event_saver.called == 0

    # Create an empty file
    test_file1.touch()

    # We might have to try several times, as filewatching can be slow
    retry = Retry()

    # New files are not considered "modified"
    assert event_saver.argument is None
    assert event_saver.called == 0

    # But when we write to it, it is considered "modified"
    test_file1.write_text('test_content')

    assert retry(lambda: event_saver.argument == test_file1)
    assert event_saver.called >= 1

    # Create a directory in the watched directory
    recursive_dir.mkdir(parents=True)

    # Subdirectories are not of interest
    assert retry(lambda: event_saver.argument == test_file1)
    assert retry(lambda: event_saver.called >= 1)

    # Create a file in the subdirectory
    test_file2.write_text('test')

    # Both the touch event and the write event are considered of interest
    assert retry(lambda: event_saver.argument == test_file2)
    assert event_saver.called == 2


def test_logging_of_os_errors(monkeypatch, tmpdir, caplog):
    """Filesystem watcher can fail due to limits, and it should be logged."""
    def raiser(self):
        raise OSError('inotify watch limit reached')

    monkeypatch.setattr(
        Observer,
        name='start',
        value=raiser,
    )

    dir_watcher = DirectoryWatcher(
        directory=tmpdir,
        on_modified=lambda x: x,
    )

    caplog.clear()
    dir_watcher.start()
    assert 'inotify watch limit reached' in caplog.record_tuples[0][2]

    dir_watcher.stop()



"""Tests for symlinking in ModuleManager."""

from astrality.module import ModuleManager


def test_symlinking_in_on_startup_block(
    action_block_factory,
    module_factory,
    create_temp_files,
):
    """ModuleManager should symlink properly."""
    file1, file2, file3, file4 = create_temp_files(4)
    file2.write_text('original')

    action_block = action_block_factory(
        symlink=[
            {'content': str(file1), 'target': str(file2)},
            {'content': str(file3), 'target': str(file4)},
        ],
    )
    module = module_factory(on_startup=action_block)

    module_manager = ModuleManager()
    module_manager.modules = {'test': module}
    module_manager.finish_tasks()

    assert file2.is_symlink()
    assert file2.resolve() == file1
    assert file4.is_symlink()
    assert file4.resolve() == file3
